import numpy as np
import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt
from sklearn.kernel_ridge import KernelRidge
from sklearn.svm import SVC
from sklearn.metrics import accuracy_score, mean_squared_error
from sklearn.model_selection import KFold, train_test_split, ParameterGrid

#from joblib import Parallel, delayed
from multiprocessing import Pool, Array
from functools import partial
import sys
import os
import time
import datetime
#from os.path import basename, splitext
from gklearn.utils.graphfiles import loadDataset
from tqdm import tqdm

#from memory_profiler import profile

#@profile
def model_selection_for_precomputed_kernel(datafile,
										   estimator,
										   param_grid_precomputed,
										   param_grid,
										   model_type,
										   NUM_TRIALS=30,
										   datafile_y=None,
										   extra_params=None,
										   ds_name='ds-unknown',
										   output_dir='outputs/',
										   n_jobs=1,
										   read_gm_from_file=False,
										   verbose=True):
	"""Perform model selection, fitting and testing for precomputed kernels 
	using nested CV. Print out neccessary data during the process then finally 
	the results.

	Parameters
	----------
	datafile : string
		Path of dataset file.
	estimator : function
		kernel function used to estimate. This function needs to return a gram matrix.
	param_grid_precomputed : dictionary
		Dictionary with names (string) of parameters used to calculate gram 
		matrices as keys and lists of parameter settings to try as values. This 
		enables searching over any sequence of parameter settings. Params with 
		length 1 will be omitted.
	param_grid : dictionary
		Dictionary with names (string) of parameters used as penelties as keys 
		and lists of parameter settings to try as values. This enables 
		searching over any sequence of parameter settings. Params with length 1
		will be omitted.
	model_type : string
		Type of the problem, can be 'regression' or 'classification'.
	NUM_TRIALS : integer
		Number of random trials of the outer CV loop. The default is 30.
	datafile_y : string
		Path of file storing y data. This parameter is optional depending on 
		the given dataset file.
	extra_params : dict
		Extra parameters for loading dataset. See function gklearn.utils.
		graphfiles.loadDataset for detail.
	ds_name : string
		Name of the dataset.
	n_jobs : int
		Number of jobs for parallelization.
	read_gm_from_file : boolean
		Whether gram matrices are loaded from a file.

	Examples
	--------
	>>> import numpy as np
	>>> from gklearn.utils.model_selection_precomputed import model_selection_for_precomputed_kernel
	>>> from gklearn.kernels.untilHPathKernel import untilhpathkernel
	>>>
	>>> datafile = '../datasets/MUTAG/MUTAG_A.txt'
	>>> estimator = untilhpathkernel
	>>> param_grid_precomputed = {’depth’:  np.linspace(1, 10, 10), ’k_func’:
			[’MinMax’, ’tanimoto’], ’compute_method’:  [’trie’]}
	>>> # ’C’ for classification problems and ’alpha’ for regression problems.
	>>> param_grid = [{’C’: np.logspace(-10, 10, num=41, base=10)}, {’alpha’:
			np.logspace(-10, 10, num=41, base=10)}]
	>>>
	>>> model_selection_for_precomputed_kernel(datafile, estimator, 
			param_grid_precomputed, param_grid[0], 'classification', ds_name=’MUTAG’)
	"""
	tqdm.monitor_interval = 0

	output_dir += estimator.__name__
	os.makedirs(output_dir, exist_ok=True)
	# a string to save all the results.
	str_fw = '###################### log time: ' + datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") + '. ######################\n\n'
	str_fw += '# This file contains results of ' + estimator.__name__ + ' on dataset ' + ds_name + ',\n# including gram matrices, serial numbers for gram matrix figures and performance.\n\n'

	# setup the model type
	model_type = model_type.lower()
	if model_type != 'regression' and model_type != 'classification':
		raise Exception(
			'The model type is incorrect! Please choose from regression or classification.'
		)
	if verbose:
		print()
		print('--- This is a %s problem ---' % model_type)
	str_fw += 'This is a %s problem.\n' % model_type
	
	# calculate gram matrices rather than read them from file.
	if read_gm_from_file == False:
		# Load the dataset
		if verbose:
			print()
			print('\n1. Loading dataset from file...')
		if isinstance(datafile, str):
			dataset, y_all = loadDataset(
					datafile, filename_y=datafile_y, extra_params=extra_params)
		else: # load data directly from variable.
			dataset = datafile
			y_all = datafile_y				

		#	 import matplotlib.pyplot as plt
		#	 import networkx as nx
		#	 nx.draw_networkx(dataset[30])
		#	 plt.show()
	
		# Grid of parameters with a discrete number of values for each.
		param_list_precomputed = list(ParameterGrid(param_grid_precomputed))
		param_list = list(ParameterGrid(param_grid))
	
		gram_matrices = [
		]  # a list to store gram matrices for all param_grid_precomputed
		gram_matrix_time = [
		]  # a list to store time to calculate gram matrices
		param_list_pre_revised = [
		]  # list to store param grids precomputed ignoring the useless ones
	
		# calculate all gram matrices
		if verbose:
			print()
			print('2. Calculating gram matrices. This could take a while...')
		str_fw += '\nII. Gram matrices.\n\n'
		tts = time.time()  # start training time
		nb_gm_ignore = 0  # the number of gram matrices those should not be considered, as they may contain elements that are not numbers (NaN)
		for idx, params_out in enumerate(param_list_precomputed):
			y = y_all[:]
			params_out['n_jobs'] = n_jobs
			params_out['verbose'] = verbose
#			print(dataset)
#			import networkx as nx
#			nx.draw_networkx(dataset[1])
#			plt.show()
			rtn_data = estimator(dataset[:], **params_out)
			Kmatrix = rtn_data[0]
			current_run_time = rtn_data[1]
			# for some kernels, some graphs in datasets may not meet the 
			# kernels' requirements for graph structure. These graphs are trimmed. 
			if len(rtn_data) == 3:
				idx_trim = rtn_data[2]  # the index of trimmed graph list
				y = [y[idxt] for idxt in idx_trim] # trim y accordingly
#			Kmatrix = np.random.rand(2250, 2250)
#			current_run_time = 0.1
			
			# remove graphs whose kernels with themselves are zeros 
			# @todo: y not changed accordingly?
			Kmatrix_diag = Kmatrix.diagonal().copy()
			nb_g_ignore = 0
			for idxk, diag in enumerate(Kmatrix_diag):
				if diag == 0:
					Kmatrix = np.delete(Kmatrix, (idxk - nb_g_ignore), axis=0)
					Kmatrix = np.delete(Kmatrix, (idxk - nb_g_ignore), axis=1)
					nb_g_ignore += 1
			# normalization
			# @todo: works only for undirected graph?
			Kmatrix_diag = Kmatrix.diagonal().copy()
			for i in range(len(Kmatrix)):
				for j in range(i, len(Kmatrix)):
					Kmatrix[i][j] /= np.sqrt(Kmatrix_diag[i] * Kmatrix_diag[j])
					Kmatrix[j][i] = Kmatrix[i][j]
			if verbose:
				print()
			if params_out == {}:
				if verbose:
					print('the gram matrix is: ')
				str_fw += 'the gram matrix is:\n\n'
			else:
				if verbose:
					print('the gram matrix with parameters', params_out, 'is: \n\n')
				str_fw += 'the gram matrix with parameters %s is:\n\n' % params_out
			if len(Kmatrix) < 2:
				nb_gm_ignore += 1
				if verbose:
					print('ignored, as at most only one of all its diagonal value is non-zero.')
				str_fw += 'ignored, as at most only one of all its diagonal value is non-zero.\n\n'
			else:				
				if np.isnan(Kmatrix).any(
				):  # if the matrix contains elements that are not numbers
					nb_gm_ignore += 1
					if verbose:
						print('ignored, as it contains elements that are not numbers.')
					str_fw += 'ignored, as it contains elements that are not numbers.\n\n'
				else:
#					print(Kmatrix)
					str_fw += np.array2string(
							Kmatrix,
							separator=',') + '\n\n'
#							separator=',',
#							threshold=np.inf,
#							floatmode='unique') + '\n\n'

					fig_file_name = output_dir + '/GM[ds]' + ds_name
					if params_out != {}:
						fig_file_name += '[params]' + str(idx)
					plt.imshow(Kmatrix)
					plt.colorbar()
					plt.savefig(fig_file_name + '.eps', format='eps', dpi=300)
#					plt.show()
					plt.clf()
					gram_matrices.append(Kmatrix)
					gram_matrix_time.append(current_run_time)
					param_list_pre_revised.append(params_out)
					if nb_g_ignore > 0:
						if verbose:
							print(', where %d graphs are ignored as their graph kernels with themselves are zeros.' % nb_g_ignore)
						str_fw += ', where %d graphs are ignored as their graph kernels with themselves are zeros.' % nb_g_ignore
		if verbose:
			print()
			print(
			'{} gram matrices are calculated, {} of which are ignored.'.format(
				len(param_list_precomputed), nb_gm_ignore))
		str_fw += '{} gram matrices are calculated, {} of which are ignored.\n\n'.format(len(param_list_precomputed), nb_gm_ignore)
		str_fw += 'serial numbers of gram matrix figures and their corresponding parameters settings:\n\n'
		str_fw += ''.join([
			'{}: {}\n'.format(idx, params_out)
			for idx, params_out in enumerate(param_list_precomputed)
		])

		if verbose:
			print()
		if len(gram_matrices) == 0:
			if verbose:
				print('all gram matrices are ignored, no results obtained.')
			str_fw += '\nall gram matrices are ignored, no results obtained.\n\n'
		else:
			# save gram matrices to file.
#			np.savez(output_dir + '/' + ds_name + '.gm', 
#					 gms=gram_matrices, params=param_list_pre_revised, y=y, 
#					 gmtime=gram_matrix_time)
			if verbose:
				print(
				'3. Fitting and predicting using nested cross validation. This could really take a while...'
				)
			
			# ---- use pool.imap_unordered to parallel and track progress. ----
#			train_pref = []
#			val_pref = []
#			test_pref = []
#			def func_assign(result, var_to_assign):
#				for idx, itm in enumerate(var_to_assign):
#					itm.append(result[idx])				
#			trial_do_partial = partial(trial_do, param_list_pre_revised, param_list, y, model_type)
#					  
#			parallel_me(trial_do_partial, range(NUM_TRIALS), func_assign, 
#						[train_pref, val_pref, test_pref], glbv=gram_matrices,
#						method='imap_unordered', n_jobs=n_jobs, chunksize=1,
#						itr_desc='cross validation')
			
			def init_worker(gms_toshare):
				global G_gms
				G_gms = gms_toshare
			
#			gram_matrices = np.array(gram_matrices)
#			gms_shape = gram_matrices.shape
#			gms_array = Array('d', np.reshape(gram_matrices.copy(), -1, order='C'))
#			pool = Pool(processes=n_jobs, initializer=init_worker, initargs=(gms_array, gms_shape))
			pool = Pool(processes=n_jobs, initializer=init_worker, initargs=(gram_matrices,))
			trial_do_partial = partial(parallel_trial_do, param_list_pre_revised, param_list, y, model_type)
			train_pref = []
			val_pref = []
			test_pref = []
#			if NUM_TRIALS < 1000 * n_jobs:
#				chunksize = int(NUM_TRIALS / n_jobs) + 1
#			else:
#				chunksize = 1000
			chunksize = 1
			if verbose:
				iterator = tqdm(pool.imap_unordered(trial_do_partial, 
						range(NUM_TRIALS), chunksize), desc='cross validation', file=sys.stdout)
			else:
				iterator = pool.imap_unordered(trial_do_partial, range(NUM_TRIALS), chunksize)
			for o1, o2, o3 in iterator:
				train_pref.append(o1)
				val_pref.append(o2)
				test_pref.append(o3)
			pool.close()
			pool.join()
	
#			# ---- use pool.map to parallel. ----
#			pool =  Pool(n_jobs)
#			trial_do_partial = partial(trial_do, param_list_pre_revised, param_list, gram_matrices, y[0:250], model_type)
#			result_perf = pool.map(trial_do_partial, range(NUM_TRIALS))
#			train_pref = [item[0] for item in result_perf]
#			val_pref = [item[1] for item in result_perf]
#			test_pref = [item[2] for item in result_perf]
	
#			# ---- direct running, normally use a single CPU core. ----
#			train_pref = []
#			val_pref = []
#			test_pref = []
#			for i in tqdm(range(NUM_TRIALS), desc='cross validation', file=sys.stdout):
#				o1, o2, o3 = trial_do(param_list_pre_revised, param_list, gram_matrices, y, model_type, i)
#				train_pref.append(o1)
#				val_pref.append(o2)
#				test_pref.append(o3)
#			print()
	
			if verbose:
				print()
				print('4. Getting final performance...')
			str_fw += '\nIII. Performance.\n\n'
			# averages and confidences of performances on outer trials for each combination of parameters
			average_train_scores = np.mean(train_pref, axis=0)
#			print('val_pref: ', val_pref[0][0])
			average_val_scores = np.mean(val_pref, axis=0)
#			print('test_pref: ', test_pref[0][0])
			average_perf_scores = np.mean(test_pref, axis=0)
			# sample std is used here
			std_train_scores = np.std(train_pref, axis=0, ddof=1)
			std_val_scores = np.std(val_pref, axis=0, ddof=1)
			std_perf_scores = np.std(test_pref, axis=0, ddof=1)
	
			if model_type == 'regression':
				best_val_perf = np.amin(average_val_scores)
			else:
				best_val_perf = np.amax(average_val_scores)
#			print('average_val_scores: ', average_val_scores)
#			print('best_val_perf: ', best_val_perf)
#			print()
			best_params_index = np.where(average_val_scores == best_val_perf)
			# find smallest val std with best val perf.
			best_val_stds = [
				std_val_scores[value][best_params_index[1][idx]]
				for idx, value in enumerate(best_params_index[0])
			]
			min_val_std = np.amin(best_val_stds)
			best_params_index = np.where(std_val_scores == min_val_std)
			best_params_out = [
				param_list_pre_revised[i] for i in best_params_index[0]
			]
			best_params_in = [param_list[i] for i in best_params_index[1]]
			if verbose:
				print('best_params_out: ', best_params_out)
				print('best_params_in: ', best_params_in)
				print()
				print('best_val_perf: ', best_val_perf)
				print('best_val_std: ', min_val_std)
			str_fw += 'best settings of hyper-params to build gram matrix: %s\n' % best_params_out
			str_fw += 'best settings of other hyper-params: %s\n\n' % best_params_in
			str_fw += 'best_val_perf: %s\n' % best_val_perf
			str_fw += 'best_val_std: %s\n' % min_val_std
	
#			print(best_params_index)
#			print(best_params_index[0])
#			print(average_perf_scores)
			final_performance = [
				average_perf_scores[value][best_params_index[1][idx]]
				for idx, value in enumerate(best_params_index[0])
			]
			final_confidence = [
				std_perf_scores[value][best_params_index[1][idx]]
				for idx, value in enumerate(best_params_index[0])
			]
			if verbose:
				print('final_performance: ', final_performance)
				print('final_confidence: ', final_confidence)
			str_fw += 'final_performance: %s\n' % final_performance
			str_fw += 'final_confidence: %s\n' % final_confidence
			train_performance = [
				average_train_scores[value][best_params_index[1][idx]]
				for idx, value in enumerate(best_params_index[0])
			]
			train_std = [
				std_train_scores[value][best_params_index[1][idx]]
				for idx, value in enumerate(best_params_index[0])
			]
			if verbose:
				print('train_performance: %s' % train_performance)
				print('train_std: ', train_std)
			str_fw += 'train_performance: %s\n' % train_performance
			str_fw += 'train_std: %s\n\n' % train_std

			if verbose:
				print()
			tt_total = time.time() - tts  # training time for all hyper-parameters
			average_gram_matrix_time = np.mean(gram_matrix_time)
			std_gram_matrix_time = np.std(gram_matrix_time, ddof=1) if len(gram_matrix_time) > 1 else 0
			best_gram_matrix_time = [
				gram_matrix_time[i] for i in best_params_index[0]
			]
			ave_bgmt = np.mean(best_gram_matrix_time)
			std_bgmt = np.std(best_gram_matrix_time, ddof=1) if len(best_gram_matrix_time) > 1 else 0
			if verbose:
				print('time to calculate gram matrix with different hyper-params: {:.2f}±{:.2f}s'
					  .format(average_gram_matrix_time, std_gram_matrix_time))
				print('time to calculate best gram matrix: {:.2f}±{:.2f}s'.format(
						ave_bgmt, std_bgmt))
				print('total training time with all hyper-param choices: {:.2f}s'.format(
						tt_total))
			str_fw += 'time to calculate gram matrix with different hyper-params: {:.2f}±{:.2f}s\n'.format(average_gram_matrix_time, std_gram_matrix_time)
			str_fw += 'time to calculate best gram matrix: {:.2f}±{:.2f}s\n'.format(ave_bgmt, std_bgmt)
			str_fw += 'total training time with all hyper-param choices: {:.2f}s\n\n'.format(tt_total)
	
			# # save results to file
			# np.savetxt(results_name_pre + 'average_train_scores.dt',
			#			average_train_scores)
			# np.savetxt(results_name_pre + 'average_val_scores', average_val_scores)
			# np.savetxt(results_name_pre + 'average_perf_scores.dt',
			#			average_perf_scores)
			# np.savetxt(results_name_pre + 'std_train_scores.dt', std_train_scores)
			# np.savetxt(results_name_pre + 'std_val_scores.dt', std_val_scores)
			# np.savetxt(results_name_pre + 'std_perf_scores.dt', std_perf_scores)
	
			# np.save(results_name_pre + 'best_params_index', best_params_index)
			# np.save(results_name_pre + 'best_params_pre.dt', best_params_out)
			# np.save(results_name_pre + 'best_params_in.dt', best_params_in)
			# np.save(results_name_pre + 'best_val_perf.dt', best_val_perf)
			# np.save(results_name_pre + 'best_val_std.dt', best_val_std)
			# np.save(results_name_pre + 'final_performance.dt', final_performance)
			# np.save(results_name_pre + 'final_confidence.dt', final_confidence)
			# np.save(results_name_pre + 'train_performance.dt', train_performance)
			# np.save(results_name_pre + 'train_std.dt', train_std)
	
			# np.save(results_name_pre + 'gram_matrix_time.dt', gram_matrix_time)
			# np.save(results_name_pre + 'average_gram_matrix_time.dt',
			#		 average_gram_matrix_time)
			# np.save(results_name_pre + 'std_gram_matrix_time.dt',
			#		 std_gram_matrix_time)
			# np.save(results_name_pre + 'best_gram_matrix_time.dt',
			#		 best_gram_matrix_time)
	
	# read gram matrices from file.
	else:	
		# Grid of parameters with a discrete number of values for each.
#		param_list_precomputed = list(ParameterGrid(param_grid_precomputed))
		param_list = list(ParameterGrid(param_grid))
	
		# read gram matrices from file.
		if verbose:
			print()
			print('2. Reading gram matrices from file...')
		str_fw += '\nII. Gram matrices.\n\nGram matrices are read from file, see last log for detail.\n'
		gmfile = np.load(output_dir + '/' + ds_name + '.gm.npz')
		gram_matrices = gmfile['gms'] # a list to store gram matrices for all param_grid_precomputed
		gram_matrix_time = gmfile['gmtime'] # time used to compute the gram matrices
		param_list_pre_revised = gmfile['params'] # list to store param grids precomputed ignoring the useless ones
		y = gmfile['y'].tolist()
		
		tts = time.time()  # start training time
#		nb_gm_ignore = 0  # the number of gram matrices those should not be considered, as they may contain elements that are not numbers (NaN)			
		if verbose:
			print(
					'3. Fitting and predicting using nested cross validation. This could really take a while...'
					)
 
		# ---- use pool.imap_unordered to parallel and track progress. ----
		def init_worker(gms_toshare):
			global G_gms
			G_gms = gms_toshare

		pool = Pool(processes=n_jobs, initializer=init_worker, initargs=(gram_matrices,))
		trial_do_partial = partial(parallel_trial_do, param_list_pre_revised, param_list, y, model_type)
		train_pref = []
		val_pref = []
		test_pref = []
		chunksize = 1
		if verbose:
			iterator = tqdm(pool.imap_unordered(trial_do_partial, 
					range(NUM_TRIALS), chunksize), desc='cross validation', file=sys.stdout)
		else:
			iterator = pool.imap_unordered(trial_do_partial, range(NUM_TRIALS), chunksize)
		for o1, o2, o3 in iterator:
			train_pref.append(o1)
			val_pref.append(o2)
			test_pref.append(o3)
		pool.close()
		pool.join()
		
		# # ---- use pool.map to parallel. ----
		# result_perf = pool.map(trial_do_partial, range(NUM_TRIALS))
		# train_pref = [item[0] for item in result_perf]
		# val_pref = [item[1] for item in result_perf]
		# test_pref = [item[2] for item in result_perf]

		# # ---- use joblib.Parallel to parallel and track progress. ----
		# trial_do_partial = partial(trial_do, param_list_pre_revised, param_list, gram_matrices, y, model_type)
		# result_perf = Parallel(n_jobs=n_jobs, verbose=10)(delayed(trial_do_partial)(trial) for trial in range(NUM_TRIALS))
		# train_pref = [item[0] for item in result_perf]
		# val_pref = [item[1] for item in result_perf]
		# test_pref = [item[2] for item in result_perf]

#		# ---- direct running, normally use a single CPU core. ----
#		train_pref = []
#		val_pref = []
#		test_pref = []
#		for i in tqdm(range(NUM_TRIALS), desc='cross validation', file=sys.stdout):
#			o1, o2, o3 = trial_do(param_list_pre_revised, param_list, gram_matrices, y, model_type, i)
#			train_pref.append(o1)
#			val_pref.append(o2)
#			test_pref.append(o3)

		if verbose:
			print()
			print('4. Getting final performance...')
		str_fw += '\nIII. Performance.\n\n'
		# averages and confidences of performances on outer trials for each combination of parameters
		average_train_scores = np.mean(train_pref, axis=0)
		average_val_scores = np.mean(val_pref, axis=0)
		average_perf_scores = np.mean(test_pref, axis=0)
		# sample std is used here
		std_train_scores = np.std(train_pref, axis=0, ddof=1)
		std_val_scores = np.std(val_pref, axis=0, ddof=1)
		std_perf_scores = np.std(test_pref, axis=0, ddof=1)

		if model_type == 'regression':
			best_val_perf = np.amin(average_val_scores)
		else:
			best_val_perf = np.amax(average_val_scores)
		best_params_index = np.where(average_val_scores == best_val_perf)
		# find smallest val std with best val perf.
		best_val_stds = [
			std_val_scores[value][best_params_index[1][idx]]
			for idx, value in enumerate(best_params_index[0])
		]
		min_val_std = np.amin(best_val_stds)
		best_params_index = np.where(std_val_scores == min_val_std)
		best_params_out = [
			param_list_pre_revised[i] for i in best_params_index[0]
		]
		best_params_in = [param_list[i] for i in best_params_index[1]]
		if verbose:
			print('best_params_out: ', best_params_out)
			print('best_params_in: ', best_params_in)
			print()
			print('best_val_perf: ', best_val_perf)
			print('best_val_std: ', min_val_std)
		str_fw += 'best settings of hyper-params to build gram matrix: %s\n' % best_params_out
		str_fw += 'best settings of other hyper-params: %s\n\n' % best_params_in
		str_fw += 'best_val_perf: %s\n' % best_val_perf
		str_fw += 'best_val_std: %s\n' % min_val_std

		final_performance = [
			average_perf_scores[value][best_params_index[1][idx]]
			for idx, value in enumerate(best_params_index[0])
		]
		final_confidence = [
			std_perf_scores[value][best_params_index[1][idx]]
			for idx, value in enumerate(best_params_index[0])
		]
		if verbose:
			print('final_performance: ', final_performance)
			print('final_confidence: ', final_confidence)
		str_fw += 'final_performance: %s\n' % final_performance
		str_fw += 'final_confidence: %s\n' % final_confidence
		train_performance = [
			average_train_scores[value][best_params_index[1][idx]]
			for idx, value in enumerate(best_params_index[0])
		]
		train_std = [
			std_train_scores[value][best_params_index[1][idx]]
			for idx, value in enumerate(best_params_index[0])
		]
		if verbose:
			print('train_performance: %s' % train_performance)
			print('train_std: ', train_std)
		str_fw += 'train_performance: %s\n' % train_performance
		str_fw += 'train_std: %s\n\n' % train_std

		if verbose:
			print()
		average_gram_matrix_time = np.mean(gram_matrix_time)
		std_gram_matrix_time = np.std(gram_matrix_time, ddof=1) if len(gram_matrix_time) > 1 else 0
		best_gram_matrix_time = [
			gram_matrix_time[i] for i in best_params_index[0]
		]
		ave_bgmt = np.mean(best_gram_matrix_time)
		std_bgmt = np.std(best_gram_matrix_time, ddof=1) if len(best_gram_matrix_time) > 1 else 0
		if verbose:		
			print(
					'time to calculate gram matrix with different hyper-params: {:.2f}±{:.2f}s'
					.format(average_gram_matrix_time, std_gram_matrix_time))
			print('time to calculate best gram matrix: {:.2f}±{:.2f}s'.format(
					ave_bgmt, std_bgmt))
		tt_poster = time.time() - tts  # training time with hyper-param choices who did not participate in calculation of gram matrices
		if verbose:
			print(
					'training time with hyper-param choices who did not participate in calculation of gram matrices: {:.2f}s'.format(
							tt_poster))
			print('total training time with all hyper-param choices: {:.2f}s'.format(
					tt_poster + np.sum(gram_matrix_time)))
#		str_fw += 'time to calculate gram matrix with different hyper-params: {:.2f}±{:.2f}s\n'.format(average_gram_matrix_time, std_gram_matrix_time)
#		str_fw += 'time to calculate best gram matrix: {:.2f}±{:.2f}s\n'.format(ave_bgmt, std_bgmt)
		str_fw += 'training time with hyper-param choices who did not participate in calculation of gram matrices: {:.2f}s\n\n'.format(tt_poster)

		# open file to save all results for this dataset.
		os.makedirs(output_dir, exist_ok=True)
			
	# print out results as table.
	str_fw += printResultsInTable(param_list, param_list_pre_revised, average_val_scores,
			  std_val_scores, average_perf_scores, std_perf_scores,
			  average_train_scores, std_train_scores, gram_matrix_time,
			  model_type, verbose)
			
	# open file to save all results for this dataset.
	if not os.path.exists(output_dir + '/' + ds_name + '.output.txt'):
		with open(output_dir + '/' + ds_name + '.output.txt', 'w') as f:
			f.write(str_fw)
	else:
		with open(output_dir + '/' + ds_name + '.output.txt', 'r+') as f:
			content = f.read()
			f.seek(0, 0)
			f.write(str_fw + '\n\n\n' + content)
			
	return final_performance, final_confidence


def trial_do(param_list_pre_revised, param_list, gram_matrices, y, model_type, trial): # Test set level

#	# get gram matrices from global variables.
#	gram_matrices = np.reshape(G_gms.copy(), G_gms_shape, order='C')
	
	# Arrays to store scores
	train_pref = np.zeros((len(param_list_pre_revised), len(param_list)))
	val_pref = np.zeros((len(param_list_pre_revised), len(param_list)))
	test_pref = np.zeros((len(param_list_pre_revised), len(param_list)))

	# randomness added to seeds of split function below. "high" is "size" times
	# 10 so that at least 10 different random output will be yielded. Remove
	# these lines if identical outputs is required.
	rdm_out = np.random.RandomState(seed=None)
	rdm_seed_out_l = rdm_out.uniform(high=len(param_list_pre_revised) * 10, 
								   size=len(param_list_pre_revised))
#	print(trial, rdm_seed_out_l)
#	print()
	# loop for each outer param tuple
	for index_out, params_out in enumerate(param_list_pre_revised):
		# get gram matrices from global variables.
#		gm_now = G_gms[index_out * G_gms_shape[1] * G_gms_shape[2]:(index_out + 1) * G_gms_shape[1] * G_gms_shape[2]]
#		gm_now = np.reshape(gm_now.copy(), (G_gms_shape[1], G_gms_shape[2]), order='C')
		gm_now = gram_matrices[index_out].copy()
	
		# split gram matrix and y to app and test sets.
		indices = range(len(y))
		# The argument "random_state" in function "train_test_split" can not be
		# set to None, because it will use RandomState instance used by 
		# np.random, which is possible for multiple subprocesses to inherit the
		# same seed if they forked at the same time, leading to identical 
		# random variates for different subprocesses. Instead, we use "trial" 
		# and "index_out" parameters to generate different seeds for different 
		# trials/subprocesses and outer loops. "rdm_seed_out_l" is used to add 
		# randomness into seeds, so that it yields a different output every 
		# time the program is run. To yield identical outputs every time,
		# remove the second line below. Same method is used to the "KFold"
		# function in the inner loop.
		rdm_seed_out = (trial + 1) * (index_out + 1)
		rdm_seed_out = (rdm_seed_out + int(rdm_seed_out_l[index_out])) % (2 ** 32 - 1)
#		print(trial, rdm_seed_out)
		X_app, X_test, y_app, y_test, idx_app, idx_test = train_test_split(
			gm_now, y, indices, test_size=0.1, 
			random_state=rdm_seed_out, shuffle=True)
#		print(trial, idx_app, idx_test)
#		print()
		X_app = X_app[:, idx_app]
		X_test = X_test[:, idx_app]
		y_app = np.array(y_app)
		y_test = np.array(y_test)

		rdm_seed_in_l = rdm_out.uniform(high=len(param_list) * 10, 
								   size=len(param_list))
		# loop for each inner param tuple
		for index_in, params_in in enumerate(param_list):
#			if trial == 0:
#				print(index_out, index_in)
#				print('params_in: ', params_in)
#			st = time.time()
			rdm_seed_in = (trial + 1) * (index_out + 1) * (index_in + 1)
#			print("rdm_seed_in1: ", trial, index_in, rdm_seed_in)
			rdm_seed_in = (rdm_seed_in + int(rdm_seed_in_l[index_in])) % (2 ** 32 - 1)
#			print("rdm_seed_in2: ", trial, index_in, rdm_seed_in)
			inner_cv = KFold(n_splits=10, shuffle=True, random_state=rdm_seed_in)
			current_train_perf = []
			current_valid_perf = []
			current_test_perf = [] 

			# For regression use the Kernel Ridge method
#			try:
			if model_type == 'regression':
				kr = KernelRidge(kernel='precomputed', **params_in)
				# loop for each split on validation set level
				# validation set level
				for train_index, valid_index in inner_cv.split(X_app):
#					print("train_index, valid_index: ", trial, index_in, train_index, valid_index)
#					if trial == 0:
#						print('train_index: ', train_index)
#						print('valid_index: ', valid_index)
#						print('idx_test: ', idx_test)
#						print('y_app[train_index]: ', y_app[train_index])
#						print('X_app[train_index, :][:, train_index]: ', X_app[train_index, :][:, train_index])
#						print('X_app[valid_index, :][:, train_index]: ', X_app[valid_index, :][:, train_index])
					kr.fit(X_app[train_index, :][:, train_index],
						   y_app[train_index])

					# predict on the train, validation and test set
					y_pred_train = kr.predict(
						X_app[train_index, :][:, train_index])
					y_pred_valid = kr.predict(
						X_app[valid_index, :][:, train_index])
#					if trial == 0:	 
#						print('y_pred_valid: ', y_pred_valid)
#						print()
					y_pred_test = kr.predict(
						X_test[:, train_index])

					# root mean squared errors
					current_train_perf.append(
						np.sqrt(
							mean_squared_error(
								y_app[train_index], y_pred_train)))
					current_valid_perf.append(
						np.sqrt(
							mean_squared_error(
								y_app[valid_index], y_pred_valid)))
#					if trial == 0:
#						print(mean_squared_error(
#								y_app[valid_index], y_pred_valid))
					current_test_perf.append(
						np.sqrt(
							mean_squared_error(
								y_test, y_pred_test)))
			# For clcassification use SVM
			else:
				svc = SVC(kernel='precomputed', cache_size=200, 
						  verbose=False, **params_in)
				# loop for each split on validation set level
				# validation set level
				for train_index, valid_index in inner_cv.split(X_app):
#						np.savez("bug.npy",X_app[train_index, :][:, train_index],y_app[train_index])
#					if trial == 0:
#						print('train_index: ', train_index)
#						print('valid_index: ', valid_index)
#						print('idx_test: ', idx_test)
#						print('y_app[train_index]: ', y_app[train_index])
#						print('X_app[train_index, :][:, train_index]: ', X_app[train_index, :][:, train_index])
#						print('X_app[valid_index, :][:, train_index]: ', X_app[valid_index, :][:, train_index])
					svc.fit(X_app[train_index, :][:, train_index],
						   y_app[train_index])
					
					# predict on the train, validation and test set
					y_pred_train = svc.predict(
						X_app[train_index, :][:, train_index])
					y_pred_valid = svc.predict(
						X_app[valid_index, :][:, train_index])
					y_pred_test = svc.predict(
						X_test[:, train_index])

					# root mean squared errors
					current_train_perf.append(
						accuracy_score(y_app[train_index],
									   y_pred_train))
					current_valid_perf.append(
						accuracy_score(y_app[valid_index],
									   y_pred_valid))
					current_test_perf.append(
						accuracy_score(y_test, y_pred_test))
#			except ValueError:
#				print(sys.exc_info()[0])
#				print(params_out, params_in)

			# average performance on inner splits
			train_pref[index_out][index_in] = np.mean(
				current_train_perf)
			val_pref[index_out][index_in] = np.mean(
				current_valid_perf)
			test_pref[index_out][index_in] = np.mean(
				current_test_perf)
#			print(time.time() - st)
#	if trial == 0:
#		print('val_pref: ', val_pref)
#		print('test_pref: ', test_pref)

	return train_pref, val_pref, test_pref

def parallel_trial_do(param_list_pre_revised, param_list, y, model_type, trial):
	train_pref, val_pref, test_pref = trial_do(param_list_pre_revised, 
											   param_list, G_gms, y, 
											   model_type, trial)
	return train_pref, val_pref, test_pref


def compute_gram_matrices(dataset, y, estimator, param_list_precomputed, 
						  output_dir, ds_name,
						  n_jobs=1, str_fw='', verbose=True):
	gram_matrices = [
		]  # a list to store gram matrices for all param_grid_precomputed
	gram_matrix_time = [
		]  # a list to store time to calculate gram matrices
	param_list_pre_revised = [
		]  # list to store param grids precomputed ignoring the useless ones
	
	nb_gm_ignore = 0  # the number of gram matrices those should not be considered, as they may contain elements that are not numbers (NaN)
	for idx, params_out in enumerate(param_list_precomputed):
		params_out['n_jobs'] = n_jobs
#			print(dataset)
#			import networkx as nx
#			nx.draw_networkx(dataset[1])
#			plt.show()
		rtn_data = estimator(dataset[:], **params_out)
		Kmatrix = rtn_data[0]
		current_run_time = rtn_data[1]
		# for some kernels, some graphs in datasets may not meet the 
		# kernels' requirements for graph structure. These graphs are trimmed. 
		if len(rtn_data) == 3:
			idx_trim = rtn_data[2]  # the index of trimmed graph list
			y = [y[idxt] for idxt in idx_trim] # trim y accordingly

		Kmatrix_diag = Kmatrix.diagonal().copy()
		# remove graphs whose kernels with themselves are zeros
		nb_g_ignore = 0
		for idxk, diag in enumerate(Kmatrix_diag):
			if diag == 0:
				Kmatrix = np.delete(Kmatrix, (idxk - nb_g_ignore), axis=0)
				Kmatrix = np.delete(Kmatrix, (idxk - nb_g_ignore), axis=1)
				nb_g_ignore += 1
		# normalization
		for i in range(len(Kmatrix)):
			for j in range(i, len(Kmatrix)):
				Kmatrix[i][j] /= np.sqrt(Kmatrix_diag[i] * Kmatrix_diag[j])
				Kmatrix[j][i] = Kmatrix[i][j]

		if verbose:
			print()
		if params_out == {}:
			if verbose:
				print('the gram matrix is: ')
			str_fw += 'the gram matrix is:\n\n'
		else:
			if verbose:
				print('the gram matrix with parameters', params_out, 'is: ')
			str_fw += 'the gram matrix with parameters %s is:\n\n' % params_out
		if len(Kmatrix) < 2:
			nb_gm_ignore += 1
			if verbose:
				print('ignored, as at most only one of all its diagonal value is non-zero.')
			str_fw += 'ignored, as at most only one of all its diagonal value is non-zero.\n\n'
		else:				
			if np.isnan(Kmatrix).any(
			):  # if the matrix contains elements that are not numbers
				nb_gm_ignore += 1
				if verbose:
					print('ignored, as it contains elements that are not numbers.')
				str_fw += 'ignored, as it contains elements that are not numbers.\n\n'
			else:
#					print(Kmatrix)
				str_fw += np.array2string(
						Kmatrix,
						separator=',') + '\n\n'
#							separator=',',
#							threshold=np.inf,
#							floatmode='unique') + '\n\n'

				fig_file_name = output_dir + '/GM[ds]' + ds_name
				if params_out != {}:
					fig_file_name += '[params]' + str(idx)
				plt.imshow(Kmatrix)
				plt.colorbar()
				plt.savefig(fig_file_name + '.eps', format='eps', dpi=300)
#					plt.show()
				plt.clf()
				gram_matrices.append(Kmatrix)
				gram_matrix_time.append(current_run_time)
				param_list_pre_revised.append(params_out)
				if nb_g_ignore > 0:
					if verbose:
						print(', where %d graphs are ignored as their graph kernels with themselves are zeros.' % nb_g_ignore)
					str_fw += ', where %d graphs are ignored as their graph kernels with themselves are zeros.' % nb_g_ignore
	if verbose:
		print()
		print(
			'{} gram matrices are calculated, {} of which are ignored.'.format(
				len(param_list_precomputed), nb_gm_ignore))
	str_fw += '{} gram matrices are calculated, {} of which are ignored.\n\n'.format(len(param_list_precomputed), nb_gm_ignore)
	str_fw += 'serial numbers of gram matrix figures and their corresponding parameters settings:\n\n'
	str_fw += ''.join([
		'{}: {}\n'.format(idx, params_out)
		for idx, params_out in enumerate(param_list_precomputed)
	])
			
	return gram_matrices, gram_matrix_time, param_list_pre_revised, y, str_fw


def read_gram_matrices_from_file(output_dir, ds_name):
	gmfile = np.load(output_dir + '/' + ds_name + '.gm.npz')
	gram_matrices = gmfile['gms'] # a list to store gram matrices for all param_grid_precomputed
	param_list_pre_revised = gmfile['params'] # list to store param grids precomputed ignoring the useless ones
	y = gmfile['y'].tolist()
	return gram_matrices, param_list_pre_revised, y


def printResultsInTable(param_list, param_list_pre_revised, average_val_scores,
						std_val_scores, average_perf_scores, std_perf_scores,
						average_train_scores, std_train_scores, gram_matrix_time,
						model_type, verbose):
	from collections import OrderedDict
	from tabulate import tabulate
	table_dict = {}
	if model_type == 'regression':
		for param_in in param_list:
			param_in['alpha'] = '{:.2e}'.format(param_in['alpha'])
	else:
		for param_in in param_list:
			param_in['C'] = '{:.2e}'.format(param_in['C'])
	table_dict['params'] = [{**param_out, **param_in}
							for param_in in param_list for param_out in param_list_pre_revised]
	table_dict['gram_matrix_time'] = [
		'{:.2f}'.format(gram_matrix_time[index_out])
		for param_in in param_list
		for index_out, _ in enumerate(param_list_pre_revised)
	]
	table_dict['valid_perf'] = [
		'{:.2f}±{:.2f}'.format(average_val_scores[index_out][index_in],
							   std_val_scores[index_out][index_in])
		for index_in, _ in enumerate(param_list)
		for index_out, _ in enumerate(param_list_pre_revised)
	]
	table_dict['test_perf'] = [
		'{:.2f}±{:.2f}'.format(average_perf_scores[index_out][index_in],
							   std_perf_scores[index_out][index_in])
		for index_in, _ in enumerate(param_list)
		for index_out, _ in enumerate(param_list_pre_revised)
	]
	table_dict['train_perf'] = [
		'{:.2f}±{:.2f}'.format(average_train_scores[index_out][index_in],
							   std_train_scores[index_out][index_in])
		for index_in, _ in enumerate(param_list)
		for index_out, _ in enumerate(param_list_pre_revised)
	]
	
	keyorder = [
		'params', 'train_perf', 'valid_perf', 'test_perf',
		'gram_matrix_time'
	]
	if verbose:
		print()
	tb_print = tabulate(OrderedDict(sorted(table_dict.items(), 
						key=lambda i: keyorder.index(i[0]))), headers='keys')
#			print(tb_print)
	return 'table of performance v.s. hyper-params:\n\n%s\n\n' % tb_print


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 12 12:52:15 2020

@author: ljia
"""
import numpy as np
import csv
import os
import os.path
from gklearn.utils import Dataset
from sklearn.model_selection import ShuffleSplit
from gklearn.preimage import MedianPreimageGenerator
from gklearn.utils import normalize_gram_matrix, compute_distance_matrix
from gklearn.preimage.utils import get_same_item_indices
from gklearn.utils.knn import knn_classification
from gklearn.preimage.utils import compute_k_dis
	

def kernel_knn_cv(ds_name, train_examples, knn_options, mpg_options, kernel_options, ged_options, mge_options, save_results=True, load_gm='auto', dir_save='', irrelevant_labels=None, edge_required=False, cut_range=None):
	
	# 1. get dataset.
	print('1. getting dataset...')
	dataset_all = Dataset()
	dataset_all.load_predefined_dataset(ds_name)
	dataset_all.trim_dataset(edge_required=edge_required)
	if irrelevant_labels is not None:
		dataset_all.remove_labels(**irrelevant_labels)
	if cut_range is not None:
		dataset_all.cut_graphs(cut_range)

	if save_results:
		# create result files.
		print('creating output files...')
		fn_output_detail, fn_output_summary = _init_output_file_knn(ds_name, kernel_options['name'], mpg_options['fit_method'], dir_save)
	else:
		fn_output_detail, fn_output_summary = None, None
		
	# 2. compute/load Gram matrix a priori.
	print('2. computing/loading Gram matrix...')
	gram_matrix_unnorm, time_precompute_gm = _get_gram_matrix(load_gm, dir_save, ds_name, kernel_options, dataset_all)
	
	# 3. perform k-nn CV.
	print('3. performing k-nn CV...')
	if train_examples == 'k-graphs' or train_examples == 'expert' or train_examples == 'random':
		_kernel_knn_cv_median(dataset_all, ds_name, knn_options, mpg_options, kernel_options, mge_options, ged_options, gram_matrix_unnorm, time_precompute_gm, train_examples, save_results, dir_save, fn_output_detail, fn_output_summary)
	
	elif train_examples == 'best-dataset':
		_kernel_knn_cv_best_ds(dataset_all, ds_name, knn_options, kernel_options, gram_matrix_unnorm, time_precompute_gm, train_examples, save_results, dir_save, fn_output_detail, fn_output_summary)
		
	elif train_examples == 'trainset':
		_kernel_knn_cv_trainset(dataset_all, ds_name, knn_options, kernel_options, gram_matrix_unnorm, time_precompute_gm, train_examples, save_results, dir_save, fn_output_detail, fn_output_summary)

	print('\ncomplete.\n')	
	
	
def _kernel_knn_cv_median(dataset_all, ds_name, knn_options, mpg_options, kernel_options, mge_options, ged_options, gram_matrix_unnorm, time_precompute_gm, train_examples, save_results, dir_save, fn_output_detail, fn_output_summary):
	Gn = dataset_all.graphs
	y_all = dataset_all.targets
	n_neighbors, n_splits, test_size = knn_options['n_neighbors'], knn_options['n_splits'], knn_options['test_size']

	# get shuffles. 
	train_indices, test_indices, train_nums, y_app = _get_shuffles(y_all, n_splits, test_size)
	
	accuracies = [[], [], []]
	for trial in range(len(train_indices)):
		print('\ntrial =', trial)
		
		train_index = train_indices[trial]
		test_index = test_indices[trial]
		G_app = [Gn[i] for i in train_index]
		G_test = [Gn[i] for i in test_index]
		y_test = [y_all[i] for i in test_index]
		gm_unnorm_trial = gram_matrix_unnorm[train_index,:][:,train_index].copy()
		
		# compute pre-images for each class.
		medians = [[], [], []]
		train_nums_tmp = [0] + train_nums
		print('\ncomputing pre-image for each class...\n')
		for i_class in range(len(train_nums_tmp) - 1):
			print(i_class + 1, 'of', len(train_nums_tmp) - 1, 'classes:')
			i_start = int(np.sum(train_nums_tmp[0:i_class + 1]))
			i_end = i_start + train_nums_tmp[i_class + 1]
			median_set = G_app[i_start:i_end]
			
			dataset = dataset_all.copy()
			dataset.load_graphs([g.copy() for g in median_set], targets=None)
			mge_options['update_order'] = True
			mpg_options['gram_matrix_unnorm'] = gm_unnorm_trial[i_start:i_end,i_start:i_end].copy()
			mpg_options['runtime_precompute_gm'] = 0
			set_median, gen_median_uo = _generate_median_preimages(dataset, mpg_options, kernel_options, ged_options, mge_options)
			mge_options['update_order'] = False
			mpg_options['gram_matrix_unnorm'] = gm_unnorm_trial[i_start:i_end,i_start:i_end].copy()
			mpg_options['runtime_precompute_gm'] = 0
			_, gen_median = _generate_median_preimages(dataset, mpg_options, kernel_options, ged_options, mge_options)
			medians[0].append(set_median)
			medians[1].append(gen_median)
			medians[2].append(gen_median_uo)
			
		# for each set of medians.
		print('\nperforming k-nn...')
		for i_app, G_app in enumerate(medians):
			# compute dis_mat between medians.
			dataset = dataset_all.copy()
			dataset.load_graphs([g.copy() for g in G_app], targets=None)
			gm_app_unnorm, _ = _compute_gram_matrix_unnorm(dataset, kernel_options.copy())
			
			# compute the entire Gram matrix.
			graph_kernel = _get_graph_kernel(dataset.copy(), kernel_options.copy())
			kernels_to_medians = []
			for g in G_app:
				kernels_to_median, _ = graph_kernel.compute(g, G_test, **kernel_options.copy()) 
				kernels_to_medians.append(kernels_to_median)
			kernels_to_medians = np.array(kernels_to_medians)
			gm_all = np.concatenate((gm_app_unnorm, kernels_to_medians), axis=1)
			gm_all = np.concatenate((gm_all, np.concatenate((kernels_to_medians.T, gram_matrix_unnorm[test_index,:][:,test_index].copy()), axis=1)), axis=0)
			
			gm_all = normalize_gram_matrix(gm_all.copy())
			dis_mat, _, _, _ = compute_distance_matrix(gm_all)
			
			N = len(G_app)
			
			d_app = dis_mat[range(N),:][:,range(N)].copy()
			
			d_test = np.zeros((N, len(test_index)))
			for i in range(N):
				for j in range(len(test_index)):
					d_test[i, j] = dis_mat[i, j]
					
			accuracies[i_app].append(knn_classification(d_app, d_test, y_app, y_test, n_neighbors, verbose=True, text=train_examples))
			
		# write result detail.
		if save_results:
			f_detail = open(dir_save + fn_output_detail, 'a')
			print('writing results to files...')
			for i, median_type in enumerate(['set-median', 'gen median', 'gen median uo']):
				csv.writer(f_detail).writerow([ds_name, kernel_options['name'], 
					  train_examples + ': ' + median_type, trial, 
					  knn_options['n_neighbors'],
					  len(gm_all), knn_options['test_size'], 
					  accuracies[i][-1][0], accuracies[i][-1][1]])
			f_detail.close()
		
	results = {}
	results['ave_perf_train'] = [np.mean([i[0] for i in j], axis=0) for j in accuracies]
	results['std_perf_train'] = [np.std([i[0] for i in j], axis=0, ddof=1) for j in accuracies]
	results['ave_perf_test'] = [np.mean([i[1] for i in j], axis=0) for j in accuracies]
	results['std_perf_test'] = [np.std([i[1] for i in j], axis=0, ddof=1) for j in accuracies]

	# write result summary for each letter. 
	if save_results:
		f_summary = open(dir_save + fn_output_summary, 'a')
		for i, median_type in enumerate(['set-median', 'gen median', 'gen median uo']):
			csv.writer(f_summary).writerow([ds_name, kernel_options['name'], 
				  train_examples + ': ' + median_type, 
				  knn_options['n_neighbors'],
				  knn_options['test_size'], results['ave_perf_train'][i],
				  results['ave_perf_test'][i], results['std_perf_train'][i],
				  results['std_perf_test'][i], time_precompute_gm])
		f_summary.close()
		
		
def _kernel_knn_cv_best_ds(dataset_all, ds_name, knn_options, kernel_options, gram_matrix_unnorm, time_precompute_gm, train_examples, save_results, dir_save, fn_output_detail, fn_output_summary):
	Gn = dataset_all.graphs
	y_all = dataset_all.targets
	n_neighbors, n_splits, test_size = knn_options['n_neighbors'], knn_options['n_splits'], knn_options['test_size']

	# get shuffles. 
	train_indices, test_indices, train_nums, y_app = _get_shuffles(y_all, n_splits, test_size)
	
	accuracies = []
	for trial in range(len(train_indices)):
		print('\ntrial =', trial)
		
		train_index = train_indices[trial]
		test_index = test_indices[trial]
		G_app = [Gn[i] for i in train_index]
		G_test = [Gn[i] for i in test_index]
		y_test = [y_all[i] for i in test_index]
		gm_unnorm_trial = gram_matrix_unnorm[train_index,:][:,train_index].copy()
		
		# get best graph from trainset according to distance in kernel space for each class.
		best_graphs = []
		train_nums_tmp = [0] + train_nums
		print('\ngetting best graph from trainset for each class...')
		for i_class in range(len(train_nums_tmp) - 1):
			print(i_class + 1, 'of', len(train_nums_tmp) - 1, 'classes.')
			i_start = int(np.sum(train_nums_tmp[0:i_class + 1]))
			i_end = i_start + train_nums_tmp[i_class + 1]
			G_class = G_app[i_start:i_end]
			gm_unnorm_class = gm_unnorm_trial[i_start:i_end,i_start:i_end]
			gm_class = normalize_gram_matrix(gm_unnorm_class.copy())
			
			k_dis_list = []
			for idx in range(len(G_class)):
				k_dis_list.append(compute_k_dis(idx, range(0, len(G_class)), [1 / len(G_class)] * len(G_class), gm_class, withterm3=False))
			idx_k_dis_min = np.argmin(k_dis_list)
			best_graphs.append(G_class[idx_k_dis_min].copy())
			
			
		# perform k-nn.
		print('\nperforming k-nn...')
		# compute dis_mat between medians.
		dataset = dataset_all.copy()
		dataset.load_graphs([g.copy() for g in best_graphs], targets=None)
		gm_app_unnorm, _ = _compute_gram_matrix_unnorm(dataset, kernel_options.copy())
		
		# compute the entire Gram matrix.
		graph_kernel = _get_graph_kernel(dataset.copy(), kernel_options.copy())
		kernels_to_best_graphs = []
		for g in best_graphs:
			kernels_to_best_graph, _ = graph_kernel.compute(g, G_test, **kernel_options.copy()) 
			kernels_to_best_graphs.append(kernels_to_best_graph)
		kernels_to_best_graphs = np.array(kernels_to_best_graphs)
		gm_all = np.concatenate((gm_app_unnorm, kernels_to_best_graphs), axis=1)
		gm_all = np.concatenate((gm_all, np.concatenate((kernels_to_best_graphs.T, gram_matrix_unnorm[test_index,:][:,test_index].copy()), axis=1)), axis=0)
		
		gm_all = normalize_gram_matrix(gm_all.copy())
		dis_mat, _, _, _ = compute_distance_matrix(gm_all)
		
		N = len(best_graphs)
		
		d_app = dis_mat[range(N),:][:,range(N)].copy()
		
		d_test = np.zeros((N, len(test_index)))
		for i in range(N):
			for j in range(len(test_index)):
				d_test[i, j] = dis_mat[i, j]
				
		accuracies.append(knn_classification(d_app, d_test, y_app, y_test, n_neighbors, verbose=True, text=train_examples))
			
		# write result detail.
		if save_results:
			f_detail = open(dir_save + fn_output_detail, 'a')
			print('writing results to files...')
			csv.writer(f_detail).writerow([ds_name, kernel_options['name'], 
				  train_examples, trial, 
				  knn_options['n_neighbors'],
				  len(gm_all), knn_options['test_size'], 
				  accuracies[-1][0], accuracies[-1][1]])
			f_detail.close()
		
	results = {}
	results['ave_perf_train'] = np.mean([i[0] for i in accuracies], axis=0)
	results['std_perf_train'] = np.std([i[0] for i in accuracies], axis=0, ddof=1)
	results['ave_perf_test'] = np.mean([i[1] for i in accuracies], axis=0)
	results['std_perf_test'] = np.std([i[1] for i in accuracies], axis=0, ddof=1)
	
	# write result summary for each letter. 
	if save_results:
		f_summary = open(dir_save + fn_output_summary, 'a')
		csv.writer(f_summary).writerow([ds_name, kernel_options['name'], 
				  train_examples, 
				  knn_options['n_neighbors'],
				  knn_options['test_size'], results['ave_perf_train'],
				  results['ave_perf_test'], results['std_perf_train'],
				  results['std_perf_test'], time_precompute_gm])
		f_summary.close()
	
	
def _kernel_knn_cv_trainset(dataset_all, ds_name, knn_options, kernel_options, gram_matrix_unnorm, time_precompute_gm, train_examples, save_results, dir_save, fn_output_detail, fn_output_summary):
	y_all = dataset_all.targets
	n_neighbors, n_splits, test_size = knn_options['n_neighbors'], knn_options['n_splits'], knn_options['test_size']
	
	# compute distance matrix.
	gram_matrix = normalize_gram_matrix(gram_matrix_unnorm.copy())
	dis_mat, _, _, _ = compute_distance_matrix(gram_matrix)

	# get shuffles. 
	train_indices, test_indices, _, _ = _get_shuffles(y_all, n_splits, test_size)
			 
	accuracies = []
	for trial in range(len(train_indices)):
		print('\ntrial =', trial)
		
		train_index = train_indices[trial]
		test_index = test_indices[trial]
		y_app = [y_all[i] for i in train_index]
		y_test = [y_all[i] for i in test_index]
		
		N = len(train_index)
		
		d_app = dis_mat[train_index,:][:,train_index].copy()
		
		d_test = np.zeros((N, len(test_index)))
		for i in range(N):
			for j in range(len(test_index)):
				d_test[i, j] = dis_mat[train_index[i], test_index[j]]
				
		accuracies.append(knn_classification(d_app, d_test, y_app, y_test, n_neighbors, verbose=True, text=train_examples))
		
		# write result detail.
		if save_results:
			print('writing results to files...')
			f_detail = open(dir_save + fn_output_detail, 'a')
			csv.writer(f_detail).writerow([ds_name, kernel_options['name'], 
					  train_examples, trial, knn_options['n_neighbors'],
					  len(gram_matrix), knn_options['test_size'], 
					  accuracies[-1][0], accuracies[-1][1]])
			f_detail.close()
		
	results = {}
	results['ave_perf_train'] = np.mean([i[0] for i in accuracies], axis=0)
	results['std_perf_train'] = np.std([i[0] for i in accuracies], axis=0, ddof=1)
	results['ave_perf_test'] = np.mean([i[1] for i in accuracies], axis=0)
	results['std_perf_test'] = np.std([i[1] for i in accuracies], axis=0, ddof=1)

	# write result summary for each letter. 
	if save_results:
		f_summary = open(dir_save + fn_output_summary, 'a')
		csv.writer(f_summary).writerow([ds_name, kernel_options['name'], 
				  train_examples, knn_options['n_neighbors'],
				  knn_options['test_size'], results['ave_perf_train'],
				  results['ave_perf_test'], results['std_perf_train'],
				  results['std_perf_test'], time_precompute_gm])
		f_summary.close()
	
	
def _get_shuffles(y_all, n_splits, test_size):
	rs = ShuffleSplit(n_splits=n_splits, test_size=test_size, random_state=0)
	train_indices = [[] for _ in range(n_splits)] 
	test_indices = [[] for _ in range(n_splits)]
	idx_targets = get_same_item_indices(y_all)
	train_nums = []
	keys = []
	for key, item in idx_targets.items():
		i = 0
		for train_i, test_i in rs.split(item): # @todo: careful when parallel.
			train_indices[i] += [item[idx] for idx in train_i]
			test_indices[i] += [item[idx] for idx in test_i]
			i += 1
		train_nums.append(len(train_i))
		keys.append(key)
	return train_indices, test_indices, train_nums, keys
	
	
def _generate_median_preimages(dataset, mpg_options, kernel_options, ged_options, mge_options):
	mpg = MedianPreimageGenerator()
	mpg.dataset = dataset.copy()
	mpg.set_options(**mpg_options.copy())
	mpg.kernel_options = kernel_options.copy()
	mpg.ged_options = ged_options.copy()
	mpg.mge_options = mge_options.copy()
	mpg.run()
	return mpg.set_median, mpg.gen_median


def _get_gram_matrix(load_gm, dir_save, ds_name, kernel_options, dataset_all):
	if load_gm == 'auto':
		gm_fname = dir_save + 'gram_matrix_unnorm.' + ds_name + '.' + kernel_options['name'] + '.gm.npz'
		gmfile_exist = os.path.isfile(os.path.abspath(gm_fname))
		if gmfile_exist:
			gmfile = np.load(gm_fname, allow_pickle=True) # @todo: may not be safe.
			gram_matrix_unnorm = gmfile['gram_matrix_unnorm']
			time_precompute_gm = float(gmfile['run_time'])
		else:
			gram_matrix_unnorm, time_precompute_gm = _compute_gram_matrix_unnorm(dataset_all, kernel_options)
			np.savez(dir_save + 'gram_matrix_unnorm.' + ds_name + '.' + kernel_options['name'] + '.gm', gram_matrix_unnorm=gram_matrix_unnorm, run_time=time_precompute_gm)
	elif not load_gm:
		gram_matrix_unnorm, time_precompute_gm = _compute_gram_matrix_unnorm(dataset_all, kernel_options)
		np.savez(dir_save + 'gram_matrix_unnorm.' + ds_name + '.' + kernel_options['name'] + '.gm', gram_matrix_unnorm=gram_matrix_unnorm, run_time=time_precompute_gm)
	else:
		gm_fname = dir_save + 'gram_matrix_unnorm.' + ds_name + '.' + kernel_options['name'] + '.gm.npz'
		gmfile = np.load(gm_fname, allow_pickle=True)
		gram_matrix_unnorm = gmfile['gram_matrix_unnorm']
		time_precompute_gm = float(gmfile['run_time'])
		
	return gram_matrix_unnorm, time_precompute_gm


def _get_graph_kernel(dataset, kernel_options):
	from gklearn.utils.utils import get_graph_kernel_by_name
	graph_kernel = get_graph_kernel_by_name(kernel_options['name'], 
				  node_labels=dataset.node_labels,
				  edge_labels=dataset.edge_labels, 
				  node_attrs=dataset.node_attrs,
				  edge_attrs=dataset.edge_attrs,
				  ds_infos=dataset.get_dataset_infos(keys=['directed']),
				  kernel_options=kernel_options)
	return graph_kernel
		
		
def _compute_gram_matrix_unnorm(dataset, kernel_options):
	from gklearn.utils.utils import get_graph_kernel_by_name
	graph_kernel = get_graph_kernel_by_name(kernel_options['name'], 
				  node_labels=dataset.node_labels,
				  edge_labels=dataset.edge_labels, 
				  node_attrs=dataset.node_attrs,
				  edge_attrs=dataset.edge_attrs,
				  ds_infos=dataset.get_dataset_infos(keys=['directed']),
				  kernel_options=kernel_options)
	
	gram_matrix, run_time = graph_kernel.compute(dataset.graphs, **kernel_options)
	gram_matrix_unnorm = graph_kernel.gram_matrix_unnorm
		
	return gram_matrix_unnorm, run_time
		
		
def _init_output_file_knn(ds_name, gkernel, fit_method, dir_output):
	if not os.path.exists(dir_output):
		os.makedirs(dir_output)
	fn_output_detail = 'results_detail_knn.' + ds_name + '.' + gkernel + '.csv'
	f_detail = open(dir_output + fn_output_detail, 'a')
	csv.writer(f_detail).writerow(['dataset', 'graph kernel',
			  'train examples', 'trial', 'num neighbors', 'num graphs', 'test size',
			  'perf train', 'perf test'])
	f_detail.close()
	
	fn_output_summary = 'results_summary_knn.' + ds_name + '.' + gkernel + '.csv'
	f_summary = open(dir_output + fn_output_summary, 'a')
	csv.writer(f_summary).writerow(['dataset', 'graph kernel',
		      'train examples', 'num neighbors', 'test size',
			  'ave perf train', 'ave perf test',
			  'std perf train', 'std perf test', 'time precompute gm'])
	f_summary.close()
	
	return fn_output_detail, fn_output_summary


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 22 15:37:36 2020

@author: ljia
"""
import numpy as np
from scipy.optimize import linear_sum_assignment


class LSAPESolver(object):
	
	
	def __init__(self, cost_matrix=None):
		"""
	/*!
	 * @brief Constructs solver for LSAPE problem instance.
	 * @param[in] cost_matrix Pointer to the LSAPE problem instance that should be solved.
	 */
		"""
		self._cost_matrix = cost_matrix
		self._model = 'ECBP'
		self._greedy_method = 'BASIC'
		self._solve_optimally = True
		self._minimal_cost = 0
		self._row_to_col_assignments = []
		self._col_to_row_assignments = []
		self._dual_var_rows = [] # @todo
		self._dual_var_cols = [] # @todo
		
	
	def clear_solution(self):
		"""Clears a previously computed solution.
		"""
		self._minimal_cost = 0
		self._row_to_col_assignments.clear()
		self._col_to_row_assignments.clear()
		self._row_to_col_assignments.append([]) # @todo
		self._col_to_row_assignments.append([])
		self._dual_var_rows = [] # @todo
		self._dual_var_cols = [] # @todo
		
		
	def set_model(self, model):
		"""
	/*!
	 * @brief Makes the solver use a specific model for optimal solving.
	 * @param[in] model The model that should be used.
	 */
		"""
		self._solve_optimally = True
		self._model = model
		
	
	def solve(self, num_solutions=1):
		"""
	/*!
	 * @brief Solves the LSAPE problem instance.
	 * @param[in] num_solutions The maximal number of solutions that should be computed.
	 */
		"""
		self.clear_solution()
		if self._solve_optimally:
			row_ind, col_ind = linear_sum_assignment(self._cost_matrix) # @todo: only hungarianLSAPE ('ECBP') can be used.
			self._row_to_col_assignments[0] = col_ind
			self._col_to_row_assignments[0] = np.argsort(col_ind) # @todo: might be slow, can use row_ind
			self._compute_cost_from_assignments()
			if num_solutions > 1:
				pass # @todo:
		else:
			print('here is non op.')
			pass # @todo: greedy.
# 			self._

	
	def minimal_cost(self):
		"""
	/*!
	 * @brief Returns the cost of the computed solutions.
	 * @return Cost of computed solutions.
	 */
		"""
		return self._minimal_cost
	
	
	def get_assigned_col(self, row, solution_id=0):
		"""
	/*!
	 * @brief Returns the assigned column.
	 * @param[in] row Row whose assigned column should be returned.
	 * @param[in] solution_id ID of the solution where the assignment should be looked up.
	 * @returns Column to which @p row is assigned to in solution with ID @p solution_id or ged::undefined() if @p row is not assigned to any column.
	 */
		"""
		return self._row_to_col_assignments[solution_id][row]
		
		
	def get_assigned_row(self, col, solution_id=0):
		"""
	/*!
	 * @brief Returns the assigned row.
	 * @param[in] col Column whose assigned row should be returned.
	 * @param[in] solution_id ID of the solution where the assignment should be looked up.
	 * @returns Row to which @p col is assigned to in solution with ID @p solution_id or ged::undefined() if @p col is not assigned to any row.
	 */
		"""
		return self._col_to_row_assignments[solution_id][col]
		
		
	def num_solutions(self):
		"""
	/*!
	 * @brief Returns the number of solutions.
	 * @returns Actual number of solutions computed by solve(). Might be smaller than @p num_solutions.
	 */
		"""
		return len(self._row_to_col_assignments)


	def _compute_cost_from_assignments(self): # @todo
		self._minimal_cost = np.sum(self._cost_matrix[range(0, len(self._row_to_col_assignments[0])), self._row_to_col_assignments[0]])


# -*- coding: utf-8 -*-
"""compute_graph_kernel_v0.1.ipynb

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/10jUz7-ahPiE_T1qvFrh2NvCVs1e47noj

**This script demonstrates how to compute a graph kernel.**
---

**0.   Install `graphkit-learn`.**
"""

"""**1.   Get dataset.**"""

from gklearn.utils.graphfiles import loadDataset

graphs, targets = loadDataset('../../../datasets/MUTAG/MUTAG_A.txt')

"""**2.  Compute graph kernel.**"""

from gklearn.kernels import untilhpathkernel

gram_matrix, run_time = untilhpathkernel(
	graphs, # The list of input graphs.
	depth=5, # The longest length of paths.
	k_func='MinMax', # Or 'tanimoto'.
	compute_method='trie', # Or 'naive'.
	n_jobs=1, # The number of jobs to run in parallel.
	verbose=True)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 31 17:06:22 2020

@author: ljia
"""
import numpy as np
from itertools import combinations
import multiprocessing
from multiprocessing import Pool
from functools import partial
import sys
# from tqdm import tqdm
import networkx as nx
from gklearn.ged.env import GEDEnv
from gklearn.utils import get_iters


def compute_ged(g1, g2, options):
	"""The simplest function to compute the graph edit distance between two graphs
	using the gedlib library.

	Notes
	-----
		- # of edit operations are not computed in this method.

	"""
	from gklearn.gedlib import librariesImport, gedlibpy

	ged_env = gedlibpy.GEDEnv()
	ged_env.set_edit_cost(
		options['edit_cost'],
		edit_cost_constant=options['edit_cost_constants']
	)
	ged_env.add_nx_graph(g1, '')
	ged_env.add_nx_graph(g2, '')
	listID = ged_env.get_all_graph_ids()
	ged_env.init(init_type=options['init_option'])
	ged_env.set_method(options['method'], ged_options_to_string(options))
	ged_env.init_method()

	g = listID[0]
	h = listID[1]
	ged_env.run_method(g, h)
	pi_forward = ged_env.get_forward_map(g, h)
	pi_backward = ged_env.get_backward_map(g, h)
	upper = ged_env.get_upper_bound(g, h)
	dis = upper

	# make the map label correct (label remove map as np.inf)
	nodes1 = [n for n in g1.nodes()]
	nodes2 = [n for n in g2.nodes()]
	nb1 = nx.number_of_nodes(g1)
	nb2 = nx.number_of_nodes(g2)
	pi_forward = [nodes2[pi] if pi < nb2 else np.inf for pi in pi_forward]
	pi_backward = [nodes1[pi] if pi < nb1 else np.inf for pi in pi_backward]
	#		print(pi_forward)

	return dis, pi_forward, pi_backward


def pairwise_ged(
		g1, g2, options={}, sort=True, repeats=1, parallel=False, verbose=True
):
	"""Compute the graph edit distance between two graphs using the gedlib library
	with repeats.

	Notes
	-----
		- For methods such as BIPARTITE, the repeats may result same results.
		- # of edit operations are not computed in this method.
	"""
	from gklearn.gedlib import librariesImport, gedlibpy

	ged_env = gedlibpy.GEDEnv()
	ged_env.set_edit_cost(
		options['edit_cost'],
		edit_cost_constant=options['edit_cost_constants']
	)
	ged_env.add_nx_graph(g1, '')
	ged_env.add_nx_graph(g2, '')
	listID = ged_env.get_all_graph_ids()
	ged_env.init(
		init_option=(options[
			             'init_option'] if 'init_option' in options else 'EAGER_WITHOUT_SHUFFLED_COPIES')
	)
	ged_env.set_method(options['method'], ged_options_to_string(options))
	ged_env.init_method()

	g = listID[0]
	h = listID[1]
	dis_min = np.inf
	# 	print('------------------------------------------')
	for i in range(0, repeats):
		ged_env.run_method(g, h)
		upper = ged_env.get_upper_bound(g, h)
		dis = upper
		# 		print(dis)
		if dis < dis_min:
			dis_min = dis
			pi_forward = ged_env.get_forward_map(g, h)
			pi_backward = ged_env.get_backward_map(g, h)
	# 			lower = ged_env.get_lower_bound(g, h)

	# make the map label correct (label remove map as np.inf)
	nodes1 = [n for n in g1.nodes()]
	nodes2 = [n for n in g2.nodes()]
	nb1 = nx.number_of_nodes(g1)
	nb2 = nx.number_of_nodes(g2)
	pi_forward = [nodes2[pi] if pi < nb2 else np.inf for pi in pi_forward]
	pi_backward = [nodes1[pi] if pi < nb1 else np.inf for pi in pi_backward]
	#		print(pi_forward)

	return dis, pi_forward, pi_backward


def compute_geds_cml(
		graphs, options={}, sort=True, parallel=False,
		verbose=True
):
	# initialize ged env.
	ged_env = GEDEnv()
	ged_env.set_edit_cost(
		options['edit_cost'],
		edit_cost_constants=options['edit_cost_constants']
	)
	for g in graphs:
		ged_env.add_nx_graph(g, '')
	listID = ged_env.get_all_graph_ids()

	node_labels = ged_env.get_all_node_labels()
	edge_labels = ged_env.get_all_edge_labels()
	node_label_costs = label_costs_to_matrix(
		options['node_label_costs'],
		len(node_labels)
	) if 'node_label_costs' in options else None
	edge_label_costs = label_costs_to_matrix(
		options['edge_label_costs'],
		len(edge_labels)
	) if 'edge_label_costs' in options else None
	ged_env.set_label_costs(node_label_costs, edge_label_costs)
	ged_env.init(init_type=options['init_option'])
	if parallel:
		options['threads'] = 1
	ged_env.set_method(options['method'], options)
	ged_env.init_method()

	# compute ged.
	# options used to compute numbers of edit operations.
	if node_label_costs is None and edge_label_costs is None:
		neo_options = {
			'edit_cost': options['edit_cost'],
			'is_cml': False,
			'node_labels': options['node_labels'],
			'edge_labels': options['edge_labels'],
			'node_attrs': options['node_attrs'],
			'edge_attrs': options['edge_attrs']
		}
	else:
		neo_options = {
			'edit_cost': options['edit_cost'],
			'is_cml': True,
			'node_labels': node_labels,
			'edge_labels': edge_labels
		}
	ged_mat = np.zeros((len(graphs), len(graphs)))
	if parallel:
		len_itr = int(len(graphs) * (len(graphs) - 1) / 2)
		ged_vec = [0 for i in range(len_itr)]
		n_edit_operations = [0 for i in range(len_itr)]
		itr = combinations(range(0, len(graphs)), 2)
		n_jobs = multiprocessing.cpu_count()
		if len_itr < 100 * n_jobs:
			chunksize = int(len_itr / n_jobs) + 1
		else:
			chunksize = 100


		def init_worker(graphs_toshare, ged_env_toshare, listID_toshare):
			global G_graphs, G_ged_env, G_listID
			G_graphs = graphs_toshare
			G_ged_env = ged_env_toshare
			G_listID = listID_toshare


		do_partial = partial(_wrapper_compute_ged_parallel, neo_options, sort)
		pool = Pool(
			processes=n_jobs, initializer=init_worker,
			initargs=(graphs, ged_env, listID)
		)
		iterator = get_iters(
			pool.imap_unordered(do_partial, itr, chunksize),
			desc='computing GEDs', file=sys.stdout,
			length=len_itr, verbose=verbose
		)
		#		iterator = pool.imap_unordered(do_partial, itr, chunksize)
		for i, j, dis, n_eo_tmp in iterator:
			idx_itr = int(len(graphs) * i + j - (i + 1) * (i + 2) / 2)
			ged_vec[idx_itr] = dis
			ged_mat[i][j] = dis
			ged_mat[j][i] = dis
			n_edit_operations[idx_itr] = n_eo_tmp
		#			print('\n-------------------------------------------')
		#			print(i, j, idx_itr, dis)
		pool.close()
		pool.join()

	else:
		ged_vec = []
		n_edit_operations = []
		iterator = get_iters(
			range(len(graphs)), desc='computing GEDs',
			file=sys.stdout, length=len(graphs),
			verbose=verbose
		)
		for i in iterator:
			#		for i in range(len(graphs)):
			for j in range(i + 1, len(graphs)):
				if nx.number_of_nodes(graphs[i]) <= nx.number_of_nodes(
						graphs[j]
				) or not sort:
					dis, pi_forward, pi_backward = _compute_ged(
						ged_env,
						listID[i],
						listID[j],
						graphs[i],
						graphs[j]
					)
				else:
					dis, pi_backward, pi_forward = _compute_ged(
						ged_env,
						listID[j],
						listID[i],
						graphs[j],
						graphs[i]
					)
				ged_vec.append(dis)
				ged_mat[i][j] = dis
				ged_mat[j][i] = dis
				n_eo_tmp = get_nb_edit_operations(
					graphs[i], graphs[j],
					pi_forward, pi_backward,
					**neo_options
				)
				n_edit_operations.append(n_eo_tmp)

	return ged_vec, ged_mat, n_edit_operations


# %%


def compute_geds(
		graphs,
		options={},
		sort=True,
		repeats=1,
		permute_nodes=False,
		random_state=None,
		parallel=False,
		n_jobs=None,
		verbose=True
):
	"""Compute graph edit distance matrix for a list of graphs using GEDLIB.

	Notes
	-----
		- With `permute_nodes=True`, the GEDs are computed by permuting the nodes a priori.
		- When `permute_nodes=False`, the GEDs can be computed with parallelization.
		- # of edit operations are computed in this method.
	"""
	if permute_nodes:
		return _compute_geds_with_permutation(
			graphs,
			options=options,
			sort=sort,
			repeats=repeats,
			random_state=random_state,
			parallel=parallel,
			n_jobs=n_jobs,
			verbose=verbose
		)
	else:
		return _compute_geds_without_permutation(
			graphs,
			options=options,
			sort=sort,
			repeats=repeats,
			parallel=parallel,
			n_jobs=n_jobs,
			verbose=verbose
		)


# %%


def _compute_geds_with_permutation(
		graphs,
		options={},
		sort=True,
		repeats=1,
		random_state=None,
		parallel=False,
		n_jobs=None,
		verbose=True
):
	"""Compute graph edit distance matrix for a list of graphs using GEDLIB, where
	the nodes are permuted a priori over different repeats.

	Notes
	-----
		- # of edit operations are computed in this method.
	"""
	from gklearn.utils.utils import nx_permute_nodes

	# Initialize variables.
	ged_mat_optim = np.full((len(graphs), len(graphs)), np.inf)
	np.fill_diagonal(ged_mat_optim, 0)
	len_itr = int(len(graphs) * (len(graphs) - 1) / 2)
	ged_vec = [0] * len_itr
	n_edit_operations = [0] * len_itr

	# for each repeat:
	for repeat in range(0, repeats):
		# Permute nodes.
		graphs_pmut = [nx_permute_nodes(g, random_state=random_state) for g in
		               graphs]

		out = _compute_geds_without_permutation(
			graphs_pmut,
			options=options,
			sort=sort,
			repeats=1,
			parallel=parallel,
			n_jobs=n_jobs,
			verbose=verbose
		)

		# Compare current results with the best one.
		idx_cnt = 0
		for i in range(len(graphs)):
			for j in range(i + 1, len(graphs)):
				if out[1][i, j] < ged_mat_optim[i, j]:
					ged_mat_optim[i, j] = out[1][i, j]
					ged_mat_optim[j, i] = out[1][j, i]
					ged_vec[idx_cnt] = out[0][idx_cnt]
					n_edit_operations[idx_cnt] = out[2][idx_cnt]
				idx_cnt += 1

	return ged_vec, ged_mat_optim, n_edit_operations


def _compute_geds_without_permutation(
		graphs,
		options={},
		sort=True,
		repeats=1,
		parallel=False,
		n_jobs=None,
		verbose=True
):
	"""Compute graph edit distance matrix for a list of graphs using GEDLIB.
	Possible parallelization with joblib, no permutation of nodes.

	Notes
	-----
		- # of edit operations are computed in this method.
	"""
	from gklearn.gedlib import librariesImport, gedlibpy

	# initialize ged env.
	ged_env = gedlibpy.GEDEnv()
	ged_env.set_edit_cost(
		options['edit_cost'],
		edit_cost_constant=options['edit_cost_constants']
	)

	for g in graphs:
		ged_env.add_nx_graph(g, '')
	listID = ged_env.get_all_graph_ids()
	ged_env.init()
	if parallel:
		options['threads'] = 1
	ged_env.set_method(options['method'], ged_options_to_string(options))
	ged_env.init_method()

	# compute ged.
	neo_options = {
		'edit_cost': options['edit_cost'],
		'node_labels': options['node_labels'],
		'edge_labels': options['edge_labels'],
		'node_attrs': options['node_attrs'],
		'edge_attrs': options['edge_attrs']
	}
	ged_mat = np.zeros((len(graphs), len(graphs)))
	if parallel:
		len_itr = int(len(graphs) * (len(graphs) - 1) / 2)
		ged_vec = [0 for i in range(len_itr)]
		n_edit_operations = [0 for i in range(len_itr)]
		itr = combinations(range(0, len(graphs)), 2)
		if n_jobs is None:
			n_jobs = multiprocessing.cpu_count()
		if len_itr < 100 * n_jobs:
			chunksize = int(len_itr / n_jobs) + 1
		else:
			chunksize = 100


		def init_worker(graphs_toshare, ged_env_toshare, listID_toshare):
			global G_graphs, G_ged_env, G_listID
			G_graphs = graphs_toshare
			G_ged_env = ged_env_toshare
			G_listID = listID_toshare


		do_partial = partial(
			_wrapper_compute_ged_parallel, neo_options, sort,
			repeats
		)
		pool = Pool(
			processes=n_jobs, initializer=init_worker,
			initargs=(graphs, ged_env, listID)
		)
		iterator = get_iters(
			pool.imap_unordered(do_partial, itr, chunksize),
			desc='computing GEDs', file=sys.stdout,
			length=len_itr, verbose=verbose
		)
		#		iterator = pool.imap_unordered(do_partial, itr, chunksize)
		for i, j, dis, n_eo_tmp in iterator:
			idx_itr = int(len(graphs) * i + j - (i + 1) * (i + 2) / 2)
			ged_vec[idx_itr] = dis
			ged_mat[i][j] = dis
			ged_mat[j][i] = dis
			n_edit_operations[idx_itr] = n_eo_tmp
		#			print('\n-------------------------------------------')
		#			print(i, j, idx_itr, dis)
		pool.close()
		pool.join()

	else:
		ged_vec = []
		n_edit_operations = []
		iterator = get_iters(
			range(len(graphs)), desc='computing GEDs',
			file=sys.stdout, length=len(graphs),
			verbose=verbose
		)
		for i in iterator:
			#		for i in range(len(graphs)):
			for j in range(i + 1, len(graphs)):
				if nx.number_of_nodes(graphs[i]) <= nx.number_of_nodes(
						graphs[j]
				) or not sort:
					dis, pi_forward, pi_backward = _compute_ged(
						ged_env,
						listID[i],
						listID[j],
						graphs[i],
						graphs[j],
						repeats
					)
				else:
					dis, pi_backward, pi_forward = _compute_ged(
						ged_env,
						listID[j],
						listID[i],
						graphs[j],
						graphs[i],
						repeats
					)
				ged_vec.append(dis)
				ged_mat[i][j] = dis
				ged_mat[j][i] = dis
				n_eo_tmp = get_nb_edit_operations(
					graphs[i], graphs[j],
					pi_forward, pi_backward,
					**neo_options
				)
				n_edit_operations.append(n_eo_tmp)

	return ged_vec, ged_mat, n_edit_operations


def _wrapper_compute_ged_parallel(options, sort, repeats, itr):
	i = itr[0]
	j = itr[1]
	dis, n_eo_tmp = _compute_ged_parallel(
		G_ged_env, G_listID[i], G_listID[j],
		G_graphs[i], G_graphs[j], options,
		sort, repeats
	)
	return i, j, dis, n_eo_tmp


def _compute_ged_parallel(env, gid1, gid2, g1, g2, options, sort, repeats):
	if nx.number_of_nodes(g1) <= nx.number_of_nodes(g2) or not sort:
		dis, pi_forward, pi_backward = _compute_ged(
			env, gid1, gid2, g1, g2,
			repeats
		)
	else:
		dis, pi_backward, pi_forward = _compute_ged(
			env, gid2, gid1, g2, g1,
			repeats
		)
	n_eo_tmp = get_nb_edit_operations(
		g1, g2, pi_forward, pi_backward,
		**options
	)  # [0,0,0,0,0,0]
	return dis, n_eo_tmp


def _compute_ged(env, gid1, gid2, g1, g2, repeats):
	dis_min = np.inf  # @todo: maybe compare distance and then do others (faster).
	for i in range(0, repeats):
		env.run_method(gid1, gid2)
		pi_forward = env.get_forward_map(gid1, gid2)
		pi_backward = env.get_backward_map(gid1, gid2)
		upper = env.get_upper_bound(gid1, gid2)
		dis = upper

		# make the map label correct (label remove map as np.inf)
		# Attention: using node indices instead of NetworkX node labels (as
		# implemented here) may cause several issues:
		# - Fail if NetworkX node labels are not consecutive integers;
		# - Return wrong mappings if nodes are permutated (e.g., by using
		# `gklearn.utis.utils.nx_permute_nodes()`.)
		nodes1 = [n for n in g1.nodes()]
		nodes2 = [n for n in g2.nodes()]
		nb1 = nx.number_of_nodes(g1)
		nb2 = nx.number_of_nodes(g2)
		pi_forward = [nodes2[pi] if pi < nb2 else np.inf for pi in pi_forward]
		pi_backward = [nodes1[pi] if pi < nb1 else np.inf for pi in pi_backward]

		if dis < dis_min:
			dis_min = dis
			pi_forward_min = pi_forward
			pi_backward_min = pi_backward

	# 	print('-----')
	# 	print(pi_forward_min)
	# 	print(pi_backward_min)

	return dis_min, pi_forward_min, pi_backward_min


# %%


def get_nb_edit_operations(
		g1, g2, forward_map, backward_map, edit_cost=None, is_cml=False,
		**kwargs
):
	"""Calculate the numbers of the occurence of each edit operation in a given
	edit path.

	Parameters
	----------
	g1 : TYPE
		DESCRIPTION.
	g2 : TYPE
		DESCRIPTION.
	forward_map : TYPE
		DESCRIPTION.
	backward_map : TYPE
		DESCRIPTION.
	edit_cost : TYPE, optional
		DESCRIPTION. The default is None.
	is_cml : TYPE, optional
		DESCRIPTION. The default is False.
	**kwargs : TYPE
		DESCRIPTION.

	Raises
	------
	Exception
		DESCRIPTION.

	Returns
	-------
	TYPE
		DESCRIPTION.

	Notes
	-----
	Attention: when implementing a function to get the numbers of edit
	operations, make sure that:
		- It does not fail if NetworkX node labels are not consecutive integers;
		- It returns correct results if nodes are permutated (e.g., by using
		`gklearn.utis.utils.nx_permute_nodes()`.)
	Generally speaking, it means you need to distinguish the NetworkX label of
	a node from the position (index) of that node in the node list.
	"""
	if is_cml:
		if edit_cost == 'CONSTANT':
			node_labels = kwargs.get('node_labels', [])
			edge_labels = kwargs.get('edge_labels', [])
			return get_nb_edit_operations_symbolic_cml(
				g1, g2, forward_map, backward_map,
				node_labels=node_labels, edge_labels=edge_labels
			)
		else:
			raise Exception('Edit cost "', edit_cost, '" is not supported.')
	else:
		if edit_cost == 'LETTER' or edit_cost == 'LETTER2':
			return get_nb_edit_operations_letter(
				g1, g2, forward_map,
				backward_map
			)
		elif edit_cost == 'NON_SYMBOLIC':
			# Use node_labels and edge_labels if node_attrs and edge_attrs are
			# not specified. Otherwise, set them to [].
			node_attrs = kwargs.get('node_attrs', kwargs.get('node_labels', []))
			edge_attrs = kwargs.get('edge_attrs', kwargs.get('edge_labels', []))
			return get_nb_edit_operations_nonsymbolic(
				g1, g2, forward_map, backward_map,
				node_attrs=node_attrs, edge_attrs=edge_attrs
			)
		elif edit_cost == 'CONSTANT':
			node_labels = kwargs.get('node_labels', [])
			edge_labels = kwargs.get('edge_labels', [])
			return get_nb_edit_operations_symbolic(
				g1, g2, forward_map, backward_map,
				node_labels=node_labels, edge_labels=edge_labels
			)
		else:
			return get_nb_edit_operations_symbolic(
				g1, g2, forward_map,
				backward_map
			)


def get_nb_edit_operations_symbolic(
		g1, g2, forward_map, backward_map, node_labels=[], edge_labels=[]
):
	"""Compute the number of each edit operations for symbolic-labeled graphs.
	"""
	n_vi = 0
	n_vr = 0
	n_vs = 0
	n_ei = 0
	n_er = 0
	n_es = 0

	nodes1 = [n for n in g1.nodes()]
	for i, map_i in enumerate(forward_map):
		if map_i == np.inf:
			n_vr += 1
		else:
			# The nodes are considered substituted as long as at least one feature
			# is different.
			for nl in node_labels:
				label1 = g1.nodes[nodes1[i]][nl]
				label2 = g2.nodes[map_i][nl]
				if label1 != label2:
					n_vs += 1
					break
	for map_i in backward_map:
		if map_i == np.inf:
			n_vi += 1

	#	idx_nodes1 = range(0, len(node1))

	edges1 = [e for e in g1.edges()]
	nb_edges2_cnted = 0
	for n1, n2 in edges1:
		idx1 = nodes1.index(n1)
		idx2 = nodes1.index(n2)
		# One of the nodes is removed, thus the edge is removed.
		if forward_map[idx1] == np.inf or forward_map[idx2] == np.inf:
			n_er += 1
		# Corresponding edge is in g2.
		elif (forward_map[idx1], forward_map[idx2]) in g2.edges():
			nb_edges2_cnted += 1
			# Edge labels are different as long as any one feature is different.
			for el in edge_labels:
				label1 = g2.edges[((forward_map[idx1], forward_map[idx2]))][el]
				label2 = g1.edges[(n1, n2)][el]
				if label1 != label2:
					n_es += 1
					break
		elif (forward_map[idx2], forward_map[idx1]) in g2.edges():
			nb_edges2_cnted += 1
			# Edge labels are different.
			for el in edge_labels:
				label1 = g2.edges[((forward_map[idx2], forward_map[idx1]))][el]
				label2 = g1.edges[(n1, n2)][el]
				if label1 != label2:
					n_es += 1
					break
		# corresponding nodes are in g2, however the edge is removed.
		else:
			n_er += 1
	n_ei = nx.number_of_edges(g2) - nb_edges2_cnted

	return n_vi, n_vr, n_vs, n_ei, n_er, n_es


def get_nb_edit_operations_symbolic_cml(
		g1, g2, forward_map, backward_map, node_labels=[], edge_labels=[]
):
	"""Compute times that edit operations are used in an edit path for symbolic-labeled graphs, where the costs are different for each pair of nodes.

	Returns
	-------
	list
		A vector of numbers of times that costs bewteen labels are used in an edit path, formed in the order of node insertion costs, node deletion costs, node substitition costs, edge insertion costs, edge deletion costs, edge substitition costs. The dummy label is the first label, and the self label costs are not included.
	"""
	# Initialize.
	nb_ops_node = np.zeros((1 + len(node_labels), 1 + len(node_labels)))
	nb_ops_edge = np.zeros((1 + len(edge_labels), 1 + len(edge_labels)))

	# For nodes.
	nodes1 = [n for n in g1.nodes()]
	for i, map_i in enumerate(forward_map):
		label1 = tuple(g1.nodes[nodes1[i]].items())  # @todo: order and faster
		idx_label1 = node_labels.index(label1)  # @todo: faster
		if map_i == np.inf:  # deletions.
			nb_ops_node[idx_label1 + 1, 0] += 1
		else:  # substitutions.
			label2 = tuple(g2.nodes[map_i].items())
			if label1 != label2:
				idx_label2 = node_labels.index(label2)  # @todo: faster
				nb_ops_node[idx_label1 + 1, idx_label2 + 1] += 1
	# insertions.
	nodes2 = [n for n in g2.nodes()]
	for i, map_i in enumerate(backward_map):
		if map_i == np.inf:
			label = tuple(g2.nodes[nodes2[i]].items())
			idx_label = node_labels.index(label)  # @todo: faster
			nb_ops_node[0, idx_label + 1] += 1

	# For edges.
	edges1 = [e for e in g1.edges()]
	edges2_marked = []
	for nf1, nt1 in edges1:
		label1 = tuple(g1.edges[(nf1, nt1)].items())
		idx_label1 = edge_labels.index(label1)  # @todo: faster
		idxf1 = nodes1.index(nf1)  # @todo: faster
		idxt1 = nodes1.index(nt1)  # @todo: faster
		# At least one of the nodes is removed, thus the edge is removed.
		if forward_map[idxf1] == np.inf or forward_map[idxt1] == np.inf:
			nb_ops_edge[idx_label1 + 1, 0] += 1
		# corresponding edge is in g2.
		else:
			nf2, nt2 = forward_map[idxf1], forward_map[idxt1]
			if (nf2, nt2) in g2.edges():
				edges2_marked.append((nf2, nt2))
				# If edge labels are different.
				label2 = tuple(g2.edges[(nf2, nt2)].items())
				if label1 != label2:
					idx_label2 = edge_labels.index(label2)  # @todo: faster
					nb_ops_edge[idx_label1 + 1, idx_label2 + 1] += 1
			# Switch nf2 and nt2, for directed graphs.
			elif (nt2, nf2) in g2.edges():
				edges2_marked.append((nt2, nf2))
				# If edge labels are different.
				label2 = tuple(g2.edges[(nt2, nf2)].items())
				if label1 != label2:
					idx_label2 = edge_labels.index(label2)  # @todo: faster
					nb_ops_edge[idx_label1 + 1, idx_label2 + 1] += 1
			# Corresponding nodes are in g2, however the edge is removed.
			else:
				nb_ops_edge[idx_label1 + 1, 0] += 1
	# insertions.
	for nt, nf in g2.edges():
		if (nt, nf) not in edges2_marked and (
				nf, nt) not in edges2_marked:  # @todo: for directed.
			label = tuple(g2.edges[(nt, nf)].items())
			idx_label = edge_labels.index(label)  # @todo: faster
			nb_ops_edge[0, idx_label + 1] += 1

	# Reform the numbers of edit oeprations into a vector.
	nb_eo_vector = []
	# node insertion.
	for i in range(1, len(nb_ops_node)):
		nb_eo_vector.append(nb_ops_node[0, i])
	# node deletion.
	for i in range(1, len(nb_ops_node)):
		nb_eo_vector.append(nb_ops_node[i, 0])
	# node substitution.
	for i in range(1, len(nb_ops_node)):
		for j in range(i + 1, len(nb_ops_node)):
			nb_eo_vector.append(nb_ops_node[i, j])
	# edge insertion.
	for i in range(1, len(nb_ops_edge)):
		nb_eo_vector.append(nb_ops_edge[0, i])
	# edge deletion.
	for i in range(1, len(nb_ops_edge)):
		nb_eo_vector.append(nb_ops_edge[i, 0])
	# edge substitution.
	for i in range(1, len(nb_ops_edge)):
		for j in range(i + 1, len(nb_ops_edge)):
			nb_eo_vector.append(nb_ops_edge[i, j])

	return nb_eo_vector


def get_nb_edit_operations_letter(g1, g2, forward_map, backward_map):
	"""Compute the number of each edit operations.
	"""
	n_vi = 0
	n_vr = 0
	n_vs = 0
	sod_vs = 0
	n_ei = 0
	n_er = 0

	nodes1 = [n for n in g1.nodes()]
	for i, map_i in enumerate(forward_map):
		if map_i == np.inf:
			n_vr += 1
		else:
			n_vs += 1
			diff_x = float(g1.nodes[nodes1[i]]['x']) - float(
				g2.nodes[map_i]['x']
			)
			diff_y = float(g1.nodes[nodes1[i]]['y']) - float(
				g2.nodes[map_i]['y']
			)
			sod_vs += np.sqrt(np.square(diff_x) + np.square(diff_y))
	for map_i in backward_map:
		if map_i == np.inf:
			n_vi += 1

	#	idx_nodes1 = range(0, len(node1))

	edges1 = [e for e in g1.edges()]
	nb_edges2_cnted = 0
	for n1, n2 in edges1:
		idx1 = nodes1.index(n1)
		idx2 = nodes1.index(n2)
		# one of the nodes is removed, thus the edge is removed.
		if forward_map[idx1] == np.inf or forward_map[idx2] == np.inf:
			n_er += 1
		# corresponding edge is in g2. Edge label is not considered.
		elif (forward_map[idx1], forward_map[idx2]) in g2.edges() or \
				(forward_map[idx2], forward_map[idx1]) in g2.edges():
			nb_edges2_cnted += 1
		# corresponding nodes are in g2, however the edge is removed.
		else:
			n_er += 1
	n_ei = nx.number_of_edges(g2) - nb_edges2_cnted

	return n_vi, n_vr, n_vs, sod_vs, n_ei, n_er


def get_nb_edit_operations_nonsymbolic(
		g1, g2, forward_map, backward_map, node_attrs=[], edge_attrs=[]
):
	"""Compute the number of each edit operations.
	"""
	n_vi = 0
	n_vr = 0
	n_vs = 0
	sod_vs = 0
	n_ei = 0
	n_er = 0
	n_es = 0
	sod_es = 0

	nodes1 = [n for n in g1.nodes()]
	for i, map_i in enumerate(forward_map):
		if map_i == np.inf:
			n_vr += 1
		else:
			n_vs += 1
			sum_squares = 0
			for a_name in node_attrs:
				diff = float(g1.nodes[nodes1[i]][a_name]) - float(
					g2.nodes[map_i][a_name]
				)
				sum_squares += np.square(diff)
			sod_vs += np.sqrt(sum_squares)
	for map_i in backward_map:
		if map_i == np.inf:
			n_vi += 1

	#	idx_nodes1 = range(0, len(node1))

	edges1 = [e for e in g1.edges()]
	for n1, n2 in edges1:
		idx1 = nodes1.index(n1)
		idx2 = nodes1.index(n2)
		n1_g2 = forward_map[idx1]
		n2_g2 = forward_map[idx2]
		# one of the nodes is removed, thus the edge is removed.
		if n1_g2 == np.inf or n2_g2 == np.inf:
			n_er += 1
		# corresponding edge is in g2.
		elif (n1_g2, n2_g2) in g2.edges():
			n_es += 1
			sum_squares = 0
			for a_name in edge_attrs:
				diff = float(g1.edges[n1, n2][a_name]) - float(
					g2.edges[n1_g2, n2_g2][a_name]
				)
				sum_squares += np.square(diff)
			sod_es += np.sqrt(sum_squares)
		elif (n2_g2, n1_g2) in g2.edges():
			n_es += 1
			sum_squares = 0
			for a_name in edge_attrs:
				diff = float(g1.edges[n2, n1][a_name]) - float(
					g2.edges[n2_g2, n1_g2][a_name]
				)
				sum_squares += np.square(diff)
			sod_es += np.sqrt(sum_squares)
		# corresponding nodes are in g2, however the edge is removed.
		else:
			n_er += 1
	n_ei = nx.number_of_edges(g2) - n_es

	return n_vi, n_vr, sod_vs, n_ei, n_er, sod_es


# %%


def label_costs_to_matrix(costs, nb_labels):
	"""Reform a label cost vector to a matrix.

	Parameters
	----------
	costs : numpy.array
		The vector containing costs between labels, in the order of node
		insertion costs, node deletion costs, node substitition costs, edge
		insertion costs, edge deletion costs, edge substitition costs.
	nb_labels : integer
		Number of labels.

	Returns
	-------
	cost_matrix : numpy.array.
		The reformed label cost matrix of size (nb_labels, nb_labels). Each
		row/column of cost_matrix corresponds to a label, and the first label
		is the dummy label. This is the same setting as in GEDData.
	"""
	# Initialize label cost matrix.
	cost_matrix = np.zeros((nb_labels + 1, nb_labels + 1))
	i = 0
	# Costs of insertions.
	for col in range(1, nb_labels + 1):
		cost_matrix[0, col] = costs[i]
		i += 1
	# Costs of deletions.
	for row in range(1, nb_labels + 1):
		cost_matrix[row, 0] = costs[i]
		i += 1
	# Costs of substitutions.
	for row in range(1, nb_labels + 1):
		for col in range(row + 1, nb_labels + 1):
			cost_matrix[row, col] = costs[i]
			cost_matrix[col, row] = costs[i]
			i += 1

	return cost_matrix


# %%


def ged_options_to_string(options):
	opt_str = ' '
	for key, val in options.items():
		if key == 'initialization_method':
			opt_str += '--initialization-method ' + str(val) + ' '
		elif key == 'initialization_options':
			opt_str += '--initialization-options ' + str(val) + ' '
		elif key == 'lower_bound_method':
			opt_str += '--lower-bound-method ' + str(val) + ' '
		elif key == 'random_substitution_ratio':
			opt_str += '--random-substitution-ratio ' + str(val) + ' '
		elif key == 'initial_solutions':
			opt_str += '--initial-solutions ' + str(val) + ' '
		elif key == 'ratio_runs_from_initial_solutions':
			opt_str += '--ratio-runs-from-initial-solutions ' + str(val) + ' '
		elif key == 'threads':
			opt_str += '--threads ' + str(val) + ' '
		elif key == 'num_randpost_loops':
			opt_str += '--num-randpost-loops ' + str(val) + ' '
		elif key == 'max_randpost_retrials':
			opt_str += '--maxrandpost-retrials ' + str(val) + ' '
		elif key == 'randpost_penalty':
			opt_str += '--randpost-penalty ' + str(val) + ' '
		elif key == 'randpost_decay':
			opt_str += '--randpost-decay ' + str(val) + ' '
		elif key == 'log':
			opt_str += '--log ' + str(val) + ' '
		elif key == 'randomness':
			opt_str += '--randomness ' + str(val) + ' '

	# 		if not isinstance(val, list):
	# 			opt_str += '--' + key.replace('_', '-') + ' '
	# 			if val == False:
	# 	 			val_str = 'FALSE'
	# 			else:
	# 	 			val_str = str(val)
	# 			opt_str += val_str + ' '

	return opt_str



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 17 15:05:01 2020

@author: ljia
"""
from gklearn.ged.env import Options, OptionsStringMap
from gklearn.ged.edit_costs import Constant
from gklearn.utils import SpecialLabel, dummy_node


class GEDData(object):
	

	def __init__(self):
		self._graphs = []
		self._graph_names = []
		self._graph_classes = []
		self._num_graphs_without_shuffled_copies = 0
		self._strings_to_internal_node_ids = []
		self._internal_node_ids_to_strings = []
		self._edit_cost = None
		self._node_costs = None
		self._edge_costs = None
		self._node_label_costs = None
		self._edge_label_costs = None
		self._node_labels = []
		self._edge_labels = []
		self._init_type = Options.InitType.EAGER_WITHOUT_SHUFFLED_COPIES
		self._delete_edit_cost = True
		self._max_num_nodes = 0
		self._max_num_edges = 0
		
		
	def num_graphs(self):
		"""
	/*!
	 * @brief Returns the number of graphs.
	 * @return Number of graphs in the instance.
	 */
		"""
		return len(self._graphs)
	
	
	def graph(self, graph_id):
		"""
	/*!
	 * @brief Provides access to a graph.
	 * @param[in] graph_id The ID of the graph.
	 * @return Constant reference to the graph with ID @p graph_id.
	 */
		"""
		return self._graphs[graph_id]
	
	
	def shuffled_graph_copies_available(self):
		"""
	/*!
	 * @brief Checks if shuffled graph copies are available.
	 * @return Boolean @p true if shuffled graph copies are available.
	 */
		"""
		return (self._init_type == Options.InitType.EAGER_WITH_SHUFFLED_COPIES or self._init_type == Options.InitType.LAZY_WITH_SHUFFLED_COPIES)
	
	
	def num_graphs_without_shuffled_copies(self):
		"""
	/*!
	 * @brief Returns the number of graphs in the instance without the shuffled copies.
	 * @return Number of graphs without shuffled copies contained in the instance.
	 */
		"""
		return self._num_graphs_without_shuffled_copies
	
	
	def node_cost(self, label1, label2):
		"""
	/*!
	 * @brief Returns node relabeling, insertion, or deletion cost.
	 * @param[in] label1 First node label.
	 * @param[in] label2 Second node label.
	 * @return Node relabeling cost if @p label1 and @p label2 are both different from ged::dummy_label(),
	 * node insertion cost if @p label1 equals ged::dummy_label and @p label2 does not,
	 * node deletion cost if @p label1 does not equal ged::dummy_label and @p label2 does,
	 * and 0 otherwise.
	 */
		"""
		if self._node_label_costs is None:
			if self._eager_init(): # @todo: check if correct
				return self._node_costs[label1, label2]
			if label1 == label2:
				return 0
			if label1 == SpecialLabel.DUMMY: # @todo: check dummy
				return self._edit_cost.node_ins_cost_fun(label2) # self._node_labels[label2 - 1]) # @todo: check
			if label2 == SpecialLabel.DUMMY: # @todo: check dummy
				return self._edit_cost.node_del_cost_fun(label1) # self._node_labels[label1 - 1])
			return self._edit_cost.node_rel_cost_fun(label1, label2) # self._node_labels[label1 - 1], self._node_labels[label2 - 1])
		# use pre-computed node label costs.
		else:
			id1 = 0 if label1 == SpecialLabel.DUMMY else self._node_label_to_id(label1) # @todo: this is slow.
			id2 = 0 if label2 == SpecialLabel.DUMMY else self._node_label_to_id(label2)
			return self._node_label_costs[id1, id2]
	
	
	def edge_cost(self, label1, label2):
		"""
	/*!
	 * @brief Returns edge relabeling, insertion, or deletion cost.
	 * @param[in] label1 First edge label.
	 * @param[in] label2 Second edge label.
	 * @return Edge relabeling cost if @p label1 and @p label2 are both different from ged::dummy_label(),
	 * edge insertion cost if @p label1 equals ged::dummy_label and @p label2 does not,
	 * edge deletion cost if @p label1 does not equal ged::dummy_label and @p label2 does,
	 * and 0 otherwise.
	 */
		"""
		if self._edge_label_costs is None:
			if self._eager_init(): # @todo: check if correct
				return self._node_costs[label1, label2]
			if label1 == label2:
				return 0
			if label1 == SpecialLabel.DUMMY:
				return self._edit_cost.edge_ins_cost_fun(label2) # self._edge_labels[label2 - 1])
			if label2 == SpecialLabel.DUMMY:
				return self._edit_cost.edge_del_cost_fun(label1) # self._edge_labels[label1 - 1])
			return self._edit_cost.edge_rel_cost_fun(label1, label2) # self._edge_labels[label1 - 1], self._edge_labels[label2 - 1])
		
		# use pre-computed edge label costs.
		else:
			id1 = 0 if label1 == SpecialLabel.DUMMY else self._edge_label_to_id(label1) # @todo: this is slow.
			id2 = 0 if label2 == SpecialLabel.DUMMY else self._edge_label_to_id(label2)
			return self._edge_label_costs[id1, id2]
	
	
	def compute_induced_cost(self, g, h, node_map):
		"""
	/*!
	 * @brief Computes the edit cost between two graphs induced by a node map.
	 * @param[in] g Input graph.
	 * @param[in] h Input graph.
	 * @param[in,out] node_map Node map whose induced edit cost is to be computed.
	 */
		"""
		cost = 0
		
		# collect node costs
		for node in g.nodes():
			image = node_map.image(node)
			label2 = (SpecialLabel.DUMMY if image == dummy_node() else h.nodes[image]['label'])
			cost += self.node_cost(g.nodes[node]['label'], label2)
		for node in h.nodes():
			pre_image = node_map.pre_image(node)
			if pre_image == dummy_node():
				cost += self.node_cost(SpecialLabel.DUMMY, h.nodes[node]['label'])
				
		# collect edge costs
		for (n1, n2) in g.edges():
			image1 = node_map.image(n1)
			image2 = node_map.image(n2)
			label2 = (h.edges[(image2, image1)]['label'] if h.has_edge(image2, image1) else SpecialLabel.DUMMY)
			cost += self.edge_cost(g.edges[(n1, n2)]['label'], label2)
		for (n1, n2) in h.edges():
			if not g.has_edge(node_map.pre_image(n2), node_map.pre_image(n1)):
				cost += self.edge_cost(SpecialLabel.DUMMY, h.edges[(n1, n2)]['label'])
				
		node_map.set_induced_cost(cost)
			
			
	def _set_edit_cost(self, edit_cost, edit_cost_constants):
		if self._delete_edit_cost:
			self._edit_cost = None
			
		if isinstance(edit_cost, str):
			edit_cost = OptionsStringMap.EditCosts[edit_cost]
			
		if edit_cost == Options.EditCosts.CHEM_1:
			if len(edit_cost_constants) == 4:
				self._edit_cost = CHEM1(edit_cost_constants[0], edit_cost_constants[1], edit_cost_constants[2], edit_cost_constants[3])
			elif len(edit_cost_constants) == 0:
				self._edit_cost = CHEM1()
			else:
				raise Exception('Wrong number of constants for selected edit costs Options::EditCosts::CHEM_1. Expected: 4 or 0; actual:', len(edit_cost_constants), '.')
		elif edit_cost == Options.EditCosts.LETTER:
			if len(edit_cost_constants) == 3:
				self._edit_cost = Letter(edit_cost_constants[0], edit_cost_constants[1], edit_cost_constants[2])
			elif len(edit_cost_constants) == 0:
				self._edit_cost = Letter()
			else:
				raise Exception('Wrong number of constants for selected edit costs Options::EditCosts::LETTER. Expected: 3 or 0; actual:', len(edit_cost_constants), '.')
		elif edit_cost == Options.EditCosts.LETTER2:
			if len(edit_cost_constants) == 5:
				self._edit_cost = Letter2(edit_cost_constants[0], edit_cost_constants[1], edit_cost_constants[2], edit_cost_constants[3], edit_cost_constants[4])
			elif len(edit_cost_constants) == 0:
				self._edit_cost = Letter2()
			else:
				raise Exception('Wrong number of constants for selected edit costs Options::EditCosts::LETTER2. Expected: 5 or 0; actual:', len(edit_cost_constants), '.')
		elif edit_cost == Options.EditCosts.NON_SYMBOLIC:
			if len(edit_cost_constants) == 6:
				self._edit_cost = NonSymbolic(edit_cost_constants[0], edit_cost_constants[1], edit_cost_constants[2], edit_cost_constants[3], edit_cost_constants[4], edit_cost_constants[5])
			elif len(edit_cost_constants) == 0:
				self._edit_cost = NonSymbolic()
			else:
				raise Exception('Wrong number of constants for selected edit costs Options::EditCosts::NON_SYMBOLIC. Expected: 6 or 0; actual:', len(edit_cost_constants), '.')
		elif edit_cost == Options.EditCosts.CONSTANT:
			if len(edit_cost_constants) == 6:
				self._edit_cost = Constant(edit_cost_constants[0], edit_cost_constants[1], edit_cost_constants[2], edit_cost_constants[3], edit_cost_constants[4], edit_cost_constants[5])
			elif len(edit_cost_constants) == 0:
				self._edit_cost = Constant()
			else:
				raise Exception('Wrong number of constants for selected edit costs Options::EditCosts::CONSTANT. Expected: 6 or 0; actual:', len(edit_cost_constants), '.')
				
		self._delete_edit_cost = True
		
		
	def id_to_node_label(self, label_id):
		if label_id > len(self._node_labels) or label_id == 0:
			raise Exception('Invalid node label ID', str(label_id), '.')
		return self._node_labels[label_id - 1]
		
		
	def _node_label_to_id(self, node_label):
		n_id = 0
		for n_l in self._node_labels:
			if n_l == node_label:
				return n_id + 1
			n_id += 1
		self._node_labels.append(node_label)
		return n_id + 1


	def id_to_edge_label(self, label_id):
		if label_id > len(self._edge_labels) or label_id == 0:
			raise Exception('Invalid edge label ID', str(label_id), '.')
		return self._edge_labels[label_id - 1]


	def _edge_label_to_id(self, edge_label):
		e_id = 0
		for e_l in self._edge_labels:
			if e_l == edge_label:
				return e_id + 1
			e_id += 1
		self._edge_labels.append(edge_label)
		return e_id + 1
		
		
	def _eager_init(self):
		return (self._init_type == Options.InitType.EAGER_WITHOUT_SHUFFLED_COPIES or self._init_type == Options.InitType.EAGER_WITH_SHUFFLED_COPIES)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Apr 13 18:02:46 2020

@author: ljia

@references:

    [1] Gaüzère B, Brun L, Villemin D. Two new graphs kernels in
    chemoinformatics. Pattern Recognition Letters. 2012 Nov 1;33(15):2038-47.
"""

import sys
from multiprocessing import Pool
from gklearn.utils import get_iters
import numpy as np
import networkx as nx
from collections import Counter
from itertools import chain
from sklearn.utils.validation import check_is_fitted
from sklearn.exceptions import NotFittedError
from gklearn.utils import SpecialLabel
from gklearn.utils.parallel import parallel_gm, parallel_me
from gklearn.utils.utils import find_all_paths, get_mlti_dim_node_attrs
from gklearn.kernels import GraphKernel


class Treelet(GraphKernel):

	def __init__(self, **kwargs):
		"""Initialise a treelet kernel.
		"""
		GraphKernel.__init__(
			self, **{
				k: kwargs.get(k) for k in
				['parallel', 'n_jobs', 'chunksize', 'normalize', 'copy_graphs',
				 'verbose'] if k in kwargs}
		)
		self.node_labels = kwargs.get('node_labels', [])
		self.edge_labels = kwargs.get('edge_labels', [])
		self.sub_kernel = kwargs.get('sub_kernel', None)
		self.ds_infos = kwargs.get('ds_infos', {})
		self.precompute_canonkeys = kwargs.get('precompute_canonkeys', True)
		self.save_canonkeys = kwargs.get('save_canonkeys', True)


	##########################################################################
	# The following is the 1st paradigm to compute kernel matrix, which is
	# compatible with `scikit-learn`.
	# -------------------------------------------------------------------
	# Special thanks to the "GraKeL" library for providing an excellent template!
	##########################################################################

	def clear_attributes(self):
		super().clear_attributes()
		if hasattr(self, '_canonkeys'):
			delattr(self, '_canonkeys')
		if hasattr(self, '_Y_canonkeys'):
			delattr(self, '_Y_canonkeys')
		if hasattr(self, '_dummy_labels_considered'):
			delattr(self, '_dummy_labels_considered')


	def validate_parameters(self):
		"""Validate all parameters for the transformer.

		Returns
		-------
		None.

		"""
		super().validate_parameters()
		if self.sub_kernel is None:
			raise ValueError('Sub-kernel not set.')


	def _compute_kernel_matrix_series(self, Y, X=None, load_canonkeys=True):
		"""Compute the kernel matrix between a given target graphs (Y) and
		the fitted graphs (X / self._graphs) without parallelization.

		Parameters
		----------
		Y : list of graphs, optional
			The target graphs.

		Returns
		-------
		kernel_matrix : numpy array, shape = [n_targets, n_inputs]
			The computed kernel matrix.

		"""
		if_comp_X_canonkeys = True

		# if load saved canonkeys of X from the instance:
		if load_canonkeys:
			# Canonical keys for self._graphs.
			try:
				check_is_fitted(self, ['_canonkeys'])
				canonkeys_list1 = self._canonkeys
				if_comp_X_canonkeys = False
			except NotFittedError:
				import warnings
				warnings.warn(
					'The canonkeys of self._graphs are not computed/saved. The keys of `X` is computed instead.'
				)
				if_comp_X_canonkeys = True

		# get all canonical keys of all graphs before computing kernels to save
		# time, but this may cost a lot of memory for large dataset.

		# Compute the canonical keys of X.
		if if_comp_X_canonkeys:
			if X is None:
				raise ('X can not be None.')
			# self._add_dummy_labels will modify the input in place.
			self._add_dummy_labels(X)  # for X
			canonkeys_list1 = []
			iterator = get_iters(
				self._graphs, desc='Getting canonkeys for X', file=sys.stdout,
				verbose=(self.verbose >= 2)
			)
			for g in iterator:
				canonkeys_list1.append(self._get_canonkeys(g))

		# Canonical keys for Y.
		# 		Y = [g.copy() for g in Y] # @todo: ?
		self._add_dummy_labels(Y)
		canonkeys_list2 = []
		iterator = get_iters(
			Y, desc='Getting canonkeys for Y', file=sys.stdout,
			verbose=(self.verbose >= 2)
		)
		for g in iterator:
			canonkeys_list2.append(self._get_canonkeys(g))

		# 		if self.save_canonkeys:
		# 			self._Y_canonkeys = canonkeys_list2

		# compute kernel matrix.
		kernel_matrix = np.zeros((len(Y), len(canonkeys_list1)))

		from itertools import product
		itr = product(range(len(Y)), range(len(canonkeys_list1)))
		len_itr = int(len(Y) * len(canonkeys_list1))
		iterator = get_iters(
			itr, desc='Computing kernels', file=sys.stdout,
			length=len_itr, verbose=(self.verbose >= 2)
		)
		for i_y, i_x in iterator:
			kernel = self._kernel_do(canonkeys_list2[i_y], canonkeys_list1[i_x])
			kernel_matrix[i_y][i_x] = kernel

		return kernel_matrix


	def _compute_kernel_matrix_imap_unordered(self, Y):
		"""Compute the kernel matrix between a given target graphs (Y) and
		the fitted graphs (X / self._graphs) using imap unordered parallelization.

		Parameters
		----------
		Y : list of graphs, optional
			The target graphs.

		Returns
		-------
		kernel_matrix : numpy array, shape = [n_targets, n_inputs]
			The computed kernel matrix.

		"""
		raise Exception('Parallelization for kernel matrix is not implemented.')


	def pairwise_kernel(self, x, y, are_keys=False):
		"""Compute pairwise kernel between two graphs.

		Parameters
		----------
		x, y : NetworkX Graph.
			Graphs bewteen which the kernel is computed.

		are_keys : boolean, optional
			If `True`, `x` and `y` are canonical keys, otherwise are graphs.
			The default is False.

		Returns
		-------
		kernel: float
			The computed kernel.

		"""
		if are_keys:
			# x, y are canonical keys.
			kernel = self._kernel_do(x, y)

		else:
			# x, y are graphs.
			kernel = self._compute_single_kernel_series(x, y)

		return kernel


	def diagonals(self):
		"""Compute the kernel matrix diagonals of the fit/transformed data.

		Returns
		-------
        X_diag : numpy array
            The diagonal of the kernel matrix between the fitted data.
            This consists of each element calculated with itself.

        Y_diag : numpy array
            The diagonal of the kernel matrix, of the transform.
            This consists of each element calculated with itself.

		"""
		# Check if method "fit" had been called.
		check_is_fitted(self, ['_graphs'])

		# Check if the diagonals of X exist.
		try:
			check_is_fitted(self, ['_X_diag'])
		except NotFittedError:
			# Compute diagonals of X.
			self._X_diag = np.empty(shape=(len(self._graphs),))
			try:
				check_is_fitted(self, ['_canonkeys'])
				for i, x in enumerate(self._canonkeys):
					self._X_diag[i] = self.pairwise_kernel(
						x, x, are_keys=True
					)  # @todo: parallel?
			except NotFittedError:
				for i, x in enumerate(self._graphs):
					self._X_diag[i] = self.pairwise_kernel(
						x, x, are_keys=False
					)  # @todo: parallel?

		try:
			# If transform has happened, return both diagonals.
			check_is_fitted(self, ['_Y'])
			self._Y_diag = np.empty(shape=(len(self._Y),))
			try:
				check_is_fitted(self, ['_Y_canonkeys'])
				for (i, y) in enumerate(self._Y_canonkeys):
					self._Y_diag[i] = self.pairwise_kernel(
						y, y, are_keys=True
					)  # @todo: parallel?
			except NotFittedError:
				for (i, y) in enumerate(self._Y):
					self._Y_diag[i] = self.pairwise_kernel(
						y, y, are_keys=False
					)  # @todo: parallel?

			return self._X_diag, self._Y_diag

		except NotFittedError:
			# Else just return both X_diag
			return self._X_diag


	##########################################################################
	# The following is the 2nd paradigm to compute kernel matrix. It is
	# simplified and not compatible with `scikit-learn`.
	##########################################################################


	def _compute_gm_series(self, graphs):
		self._add_dummy_labels(graphs)

		# get all canonical keys of all graphs before computing kernels to save
		# time, but this may cost a lot of memory for large dataset.
		canonkeys = []
		iterator = get_iters(
			graphs, desc='getting canonkeys', file=sys.stdout,
			verbose=(self.verbose >= 2)
		)
		for g in iterator:
			canonkeys.append(self._get_canonkeys(g))

		if self.save_canonkeys:
			self._canonkeys = canonkeys

		# compute Gram matrix.
		gram_matrix = np.zeros((len(graphs), len(graphs)))

		from itertools import combinations_with_replacement
		itr = combinations_with_replacement(range(0, len(graphs)), 2)
		len_itr = int(len(graphs) * (len(graphs) + 1) / 2)
		iterator = get_iters(
			itr, desc='Computing kernels', file=sys.stdout,
			length=len_itr, verbose=(self.verbose >= 2)
		)
		for i, j in iterator:
			kernel = self._kernel_do(canonkeys[i], canonkeys[j])
			gram_matrix[i][j] = kernel
			gram_matrix[j][i] = kernel  # @todo: no directed graph considered?

		return gram_matrix


	def _compute_gm_imap_unordered(self):
		self._add_dummy_labels(self._graphs)

		# get all canonical keys of all graphs before computing kernels to save
		# time, but this may cost a lot of memory for large dataset.
		pool = Pool(self.n_jobs)
		itr = zip(self._graphs, range(0, len(self._graphs)))
		if len(self._graphs) < 100 * self.n_jobs:
			chunksize = int(len(self._graphs) / self.n_jobs) + 1
		else:
			chunksize = 100
		canonkeys = [[] for _ in range(len(self._graphs))]
		get_fun = self._wrapper_get_canonkeys
		iterator = get_iters(
			pool.imap_unordered(get_fun, itr, chunksize),
			desc='getting canonkeys', file=sys.stdout,
			length=len(self._graphs), verbose=(self.verbose >= 2)
		)
		for i, ck in iterator:
			canonkeys[i] = ck
		pool.close()
		pool.join()

		if self.save_canonkeys:
			self._canonkeys = canonkeys

		# compute Gram matrix.
		gram_matrix = np.zeros((len(self._graphs), len(self._graphs)))


		def init_worker(canonkeys_toshare):
			global G_canonkeys
			G_canonkeys = canonkeys_toshare


		do_fun = self._wrapper_kernel_do
		parallel_gm(
			do_fun, gram_matrix, self._graphs, init_worker=init_worker,
			glbv=(canonkeys,), n_jobs=self.n_jobs, verbose=self.verbose
		)

		return gram_matrix


	def _compute_kernel_list_series(self, g1, g_list):
		# @TODO: Why this is commented out before?
		self._add_dummy_labels(g_list + [g1])

		# get all canonical keys of all graphs before computing kernels to save
		# time, but this may cost a lot of memory for large dataset.
		canonkeys_1 = self._get_canonkeys(g1)
		canonkeys_list = []
		iterator = get_iters(
			g_list, desc='getting canonkeys', file=sys.stdout,
			verbose=(self.verbose >= 2)
		)
		for g in iterator:
			canonkeys_list.append(self._get_canonkeys(g))

		# compute kernel list.
		kernel_list = [None] * len(g_list)
		iterator = get_iters(
			range(len(g_list)), desc='Computing kernels', file=sys.stdout,
			length=len(g_list), verbose=(self.verbose >= 2)
		)
		for i in iterator:
			kernel = self._kernel_do(canonkeys_1, canonkeys_list[i])
			kernel_list[i] = kernel

		return kernel_list


	def _compute_kernel_list_imap_unordered(self, g1, g_list):
		self._add_dummy_labels(g_list + [g1])

		# get all canonical keys of all graphs before computing kernels to save
		# time, but this may cost a lot of memory for large dataset.
		canonkeys_1 = self._get_canonkeys(g1)
		canonkeys_list = [[] for _ in range(len(g_list))]
		pool = Pool(self.n_jobs)
		itr = zip(g_list, range(0, len(g_list)))
		if len(g_list) < 100 * self.n_jobs:
			chunksize = int(len(g_list) / self.n_jobs) + 1
		else:
			chunksize = 100
		get_fun = self._wrapper_get_canonkeys
		iterator = get_iters(
			pool.imap_unordered(get_fun, itr, chunksize),
			desc='getting canonkeys', file=sys.stdout,
			length=len(g_list), verbose=(self.verbose >= 2)
		)
		for i, ck in iterator:
			canonkeys_list[i] = ck
		pool.close()
		pool.join()

		# compute kernel list.
		kernel_list = [None] * len(g_list)


		def init_worker(ck_1_toshare, ck_list_toshare):
			global G_ck_1, G_ck_list
			G_ck_1 = ck_1_toshare
			G_ck_list = ck_list_toshare


		do_fun = self._wrapper_kernel_list_do


		def func_assign(result, var_to_assign):
			var_to_assign[result[0]] = result[1]


		itr = range(len(g_list))
		len_itr = len(g_list)
		parallel_me(
			do_fun, func_assign, kernel_list, itr, len_itr=len_itr,
			init_worker=init_worker, glbv=(canonkeys_1, canonkeys_list),
			method='imap_unordered',
			n_jobs=self.n_jobs, itr_desc='Computing kernels',
			verbose=self.verbose
		)

		return kernel_list


	def _wrapper_kernel_list_do(self, itr):
		return itr, self._kernel_do(G_ck_1, G_ck_list[itr])


	def _compute_single_kernel_series(self, g1, g2):
		# @TODO: Why this is commented out before?
		self._add_dummy_labels([g1] + [g2])
		canonkeys_1 = self._get_canonkeys(g1)
		canonkeys_2 = self._get_canonkeys(g2)
		kernel = self._kernel_do(canonkeys_1, canonkeys_2)
		return kernel


	# 	@profile
	def _kernel_do(self, canonkey1, canonkey2):
		"""Compute treelet graph kernel between 2 graphs.

		Parameters
		----------
		canonkey1, canonkey2 : list
			List of canonical keys in 2 graphs, where each key is represented by a string.

		Return
		------
		kernel : float
			Treelet kernel between 2 graphs.
		"""
		keys = set(canonkey1.keys()) | set(canonkey2.keys())  # find same canonical keys in both graphs
		if len(keys) == 0:  # There is nothing in common...
		        return 0

		vector1 = np.array([canonkey1.get(key, 0) for key in keys])
		vector2 = np.array([canonkey2.get(key, 0) for key in keys])

		# 		vector1, vector2 = [], []
		# 		keys1, keys2 = canonkey1, canonkey2
		# 		keys_searched = {}
		# 		for k, v in canonkey1.items():
		# 			if k in keys2:
		# 				vector1.append(v)
		# 				vector2.append(canonkey2[k])
		# 				keys_searched[k] = v

		# 		for k, v in canonkey2.items():
		# 			if k in keys1 and k not in keys_searched:
		# 				vector1.append(canonkey1[k])
		# 				vector2.append(v)

		# 		vector1, vector2 = np.array(vector1), np.array(vector2)

		kernel = self.sub_kernel(vector1, vector2)
		return kernel


	def _wrapper_kernel_do(self, itr):
		i = itr[0]
		j = itr[1]
		return i, j, self._kernel_do(G_canonkeys[i], G_canonkeys[j])


	def _get_canonkeys(self, G):
		"""Generate canonical keys of all treelets in a graph.

		Parameters
		----------
		G : NetworkX graphs
			The graph in which keys are generated.

		Return
		------
		canonkey/canonkey_l : dict
			For unlabeled graphs, canonkey is a dictionary which records amount of
			every tree pattern. For labeled graphs, canonkey_l is one which keeps
			track of amount of every treelet.
		"""
		patterns = {}  # a dictionary which consists of lists of patterns for all graphlet.
		canonkey = {}  # canonical key, a dictionary which records amount of every tree pattern.

		### structural analysis ###
		### In this section, a list of patterns is generated for each graphlet,
		### where every pattern is represented by nodes ordered by Morgan's
		### extended labeling.
		# linear patterns
		patterns['0'] = list(G.nodes())
		canonkey['0'] = nx.number_of_nodes(G)
		for i in range(1, 6):  # for i in range(1, 6):
			patterns[str(i)] = find_all_paths(G, i, self.ds_infos['directed'])
			canonkey[str(i)] = len(patterns[str(i)])

		# n-star patterns
		patterns['3star'] = [[node] + [neighbor for neighbor in G[node]] for
		                     node in G.nodes() if len(G[node]) == 3]
		patterns['4star'] = [[node] + [neighbor for neighbor in G[node]] for
		                     node in G.nodes() if len(G[node]) == 4]
		patterns['5star'] = [[node] + [neighbor for neighbor in G[node]] for
		                     node in G.nodes() if len(G[node]) == 5]
		# n-star patterns
		canonkey['6'] = len(patterns['3star'])
		canonkey['8'] = len(patterns['4star'])
		canonkey['d'] = len(patterns['5star'])

		# pattern 7
		patterns['7'] = []  # the 1st line of Table 1 in Ref [1]
		for pattern in patterns['3star']:
			for i in range(1, len(pattern)):  # for each neighbor of node 0
				if len(G[pattern[i]]) >= 2:
					pattern_t = pattern[:]
					# set the node with degree >= 2 as the 4th node
					pattern_t[i], pattern_t[3] = pattern_t[3], pattern_t[i]
					for neighborx in G[pattern[i]]:
						if neighborx != pattern[0]:
							new_pattern = pattern_t + [neighborx]
							patterns['7'].append(new_pattern)
		canonkey['7'] = len(patterns['7'])

		# pattern 11
		patterns['11'] = []  # the 4th line of Table 1 in Ref [1]
		for pattern in patterns['4star']:
			for i in range(1, len(pattern)):
				if len(G[pattern[i]]) >= 2:
					pattern_t = pattern[:]
					pattern_t[i], pattern_t[4] = pattern_t[4], pattern_t[i]
					for neighborx in G[pattern[i]]:
						if neighborx != pattern[0]:
							new_pattern = pattern_t + [neighborx]
							patterns['11'].append(new_pattern)
		canonkey['b'] = len(patterns['11'])

		# pattern 12
		patterns['12'] = []  # the 5th line of Table 1 in Ref [1]
		rootlist = []  # a list of root nodes, whose extended labels are 3
		for pattern in patterns['3star']:
			if pattern[
				0] not in rootlist:  # prevent to count the same pattern twice from each of the two root nodes
				rootlist.append(pattern[0])
				for i in range(1, len(pattern)):
					if len(G[pattern[i]]) >= 3:
						rootlist.append(pattern[i])
						pattern_t = pattern[:]
						pattern_t[i], pattern_t[3] = pattern_t[3], pattern_t[i]
						for neighborx1 in G[pattern[i]]:
							if neighborx1 != pattern[0]:
								for neighborx2 in G[pattern[i]]:
									if neighborx1 > neighborx2 and neighborx2 != \
											pattern[0]:
										new_pattern = pattern_t + [
											neighborx1] + [neighborx2]
										#						 new_patterns = [ pattern + [neighborx1] + [neighborx2] for neighborx1 in G[pattern[i]] if neighborx1 != pattern[0] for neighborx2 in G[pattern[i]] if (neighborx1 > neighborx2 and neighborx2 != pattern[0]) ]
										patterns['12'].append(new_pattern)
		canonkey['c'] = int(len(patterns['12']) / 2)

		# pattern 9
		# todo: this is not correct for self loops, but for now, we simply remove
		# self loops from the graph at the beginning of the model,
		# see GraphKernel._compute_gram_matrix().
		patterns['9'] = []  # the 2nd line of Table 1 in Ref [1]
		for pattern in patterns['3star']:
			for pairs in [[neighbor1, neighbor2] for neighbor1 in G[pattern[0]]
			              if len(G[neighbor1]) >= 2 \
			              for neighbor2 in G[pattern[0]] if
			              len(G[neighbor2]) >= 2 if neighbor1 > neighbor2]:
				pattern_t = pattern[:]
				# move nodes with extended labels 4 to specific position to correspond to their children
				pattern_t[pattern_t.index(pairs[0])], pattern_t[2] = pattern_t[
					2], pattern_t[pattern_t.index(pairs[0])]
				pattern_t[pattern_t.index(pairs[1])], pattern_t[3] = pattern_t[
					3], pattern_t[pattern_t.index(pairs[1])]
				for neighborx1 in G[pairs[0]]:
					if neighborx1 != pattern[0]:
						for neighborx2 in G[pairs[1]]:
							if neighborx2 != pattern[0]:
								new_pattern = pattern_t + [neighborx1] + [
									neighborx2]
								patterns['9'].append(new_pattern)
		canonkey['9'] = len(patterns['9'])

		# pattern 10
		patterns['10'] = []  # the 3rd line of Table 1 in Ref [1]
		for pattern in patterns['3star']:
			for i in range(1, len(pattern)):
				if len(G[pattern[i]]) >= 2:
					for neighborx in G[pattern[i]]:
						if neighborx != pattern[0] and len(G[neighborx]) >= 2:
							pattern_t = pattern[:]
							pattern_t[i], pattern_t[3] = pattern_t[3], \
							pattern_t[i]
							new_patterns = [
								pattern_t + [neighborx] + [neighborxx] for
								neighborxx in G[neighborx] if
								neighborxx != pattern[i]]
							patterns['10'].extend(new_patterns)
		canonkey['a'] = len(patterns['10'])

		### labeling information ###
		### In this section, a list of canonical keys is generated for every
		### pattern obtained in the structural analysis section above, which is a
		### string corresponding to a unique treelet. A dictionary is built to keep
		### track of the amount of every treelet.
		if len(self.node_labels) > 0 or len(self.edge_labels) > 0:
			canonkey_l = {}  # canonical key, a dictionary which keeps track of amount of every treelet.

			# linear patterns
			canonkey_t = Counter(get_mlti_dim_node_attrs(G, self.node_labels))
			for key in canonkey_t:
				canonkey_l[('0', key)] = canonkey_t[key]

			for i in range(1, 6):  # for i in range(1, 6):
				treelet = []
				for pattern in patterns[str(i)]:
					canonlist = []
					for idx, node in enumerate(pattern[:-1]):
						canonlist.append(
							tuple(G.nodes[node][nl] for nl in self.node_labels)
						)
						canonlist.append(
							tuple(
								G[node][pattern[idx + 1]][el] for el in
								self.edge_labels
							)
						)
					canonlist.append(
						tuple(
							G.nodes[pattern[-1]][nl] for nl in self.node_labels
						)
					)
					canonkey_t = canonlist if canonlist < canonlist[
					                                      ::-1] else canonlist[
					                                                 ::-1]
					treelet.append(tuple([str(i)] + canonkey_t))
				canonkey_l.update(Counter(treelet))

			# n-star patterns
			for i in range(3, 6):
				treelet = []
				for pattern in patterns[str(i) + 'star']:
					canonlist = []
					for leaf in pattern[1:]:
						nlabels = tuple(
							G.nodes[leaf][nl] for nl in self.node_labels
						)
						elabels = tuple(
							G[leaf][pattern[0]][el] for el in self.edge_labels
						)
						canonlist.append(tuple((nlabels, elabels)))
					canonlist.sort()
					canonlist = list(chain.from_iterable(canonlist))
					canonkey_t = tuple(
						['d' if i == 5 else str(i * 2)] +
						[tuple(
							G.nodes[pattern[0]][nl] for nl in self.node_labels
						)]
						+ canonlist
						)
					treelet.append(canonkey_t)
				canonkey_l.update(Counter(treelet))

			# pattern 7
			treelet = []
			for pattern in patterns['7']:
				canonlist = []
				for leaf in pattern[1:3]:
					nlabels = tuple(
						G.nodes[leaf][nl] for nl in self.node_labels
					)
					elabels = tuple(
						G[leaf][pattern[0]][el] for el in self.edge_labels
					)
					canonlist.append(tuple((nlabels, elabels)))
				canonlist.sort()
				canonlist = list(chain.from_iterable(canonlist))
				canonkey_t = tuple(
					['7']
					+ [tuple(
						G.nodes[pattern[0]][nl] for nl in self.node_labels
					)] + canonlist
					+ [tuple(
						G.nodes[pattern[3]][nl] for nl in self.node_labels
					)]
					+ [tuple(
						G[pattern[3]][pattern[0]][el] for el in self.edge_labels
					)]
					+ [tuple(
						G.nodes[pattern[4]][nl] for nl in self.node_labels
					)]
					+ [tuple(
						G[pattern[4]][pattern[3]][el] for el in self.edge_labels
					)]
					)
				treelet.append(canonkey_t)
			canonkey_l.update(Counter(treelet))

			# pattern 11
			treelet = []
			for pattern in patterns['11']:
				canonlist = []
				for leaf in pattern[1:4]:
					nlabels = tuple(
						G.nodes[leaf][nl] for nl in self.node_labels
					)
					elabels = tuple(
						G[leaf][pattern[0]][el] for el in self.edge_labels
					)
					canonlist.append(tuple((nlabels, elabels)))
				canonlist.sort()
				canonlist = list(chain.from_iterable(canonlist))
				canonkey_t = tuple(
					['b']
					+ [tuple(
						G.nodes[pattern[0]][nl] for nl in self.node_labels
					)] + canonlist
					+ [tuple(
						G.nodes[pattern[4]][nl] for nl in self.node_labels
					)]
					+ [tuple(
						G[pattern[4]][pattern[0]][el] for el in self.edge_labels
					)]
					+ [tuple(
						G.nodes[pattern[5]][nl] for nl in self.node_labels
					)]
					+ [tuple(
						G[pattern[5]][pattern[4]][el] for el in self.edge_labels
					)]
					)
				treelet.append(canonkey_t)
			canonkey_l.update(Counter(treelet))

			# pattern 10
			treelet = []
			for pattern in patterns['10']:
				canonkey4 = [
					tuple(G.nodes[pattern[5]][nl] for nl in self.node_labels),
					tuple(
						G[pattern[5]][pattern[4]][el] for el in self.edge_labels
					)]
				canonlist = []
				for leaf in pattern[1:3]:
					nlabels = tuple(
						G.nodes[leaf][nl] for nl in self.node_labels
					)
					elabels = tuple(
						G[leaf][pattern[0]][el] for el in self.edge_labels
					)
					canonlist.append(tuple((nlabels, elabels)))
				canonlist.sort()
				canonkey0 = list(chain.from_iterable(canonlist))
				canonkey_t = tuple(
					['a']
					+ [tuple(
						G.nodes[pattern[3]][nl] for nl in self.node_labels
					)]
					+ [tuple(
						G.nodes[pattern[4]][nl] for nl in self.node_labels
					)]
					+ [tuple(
						G[pattern[4]][pattern[3]][el] for el in self.edge_labels
					)]
					+ [tuple(
						G.nodes[pattern[0]][nl] for nl in self.node_labels
					)]
					+ [tuple(
						G[pattern[0]][pattern[3]][el] for el in self.edge_labels
					)]
					+ canonkey4 + canonkey0
					)
				treelet.append(canonkey_t)
			canonkey_l.update(Counter(treelet))

			# pattern 12
			treelet = []
			for pattern in patterns['12']:
				canonlist0 = []
				for leaf in pattern[1:3]:
					nlabels = tuple(
						G.nodes[leaf][nl] for nl in self.node_labels
					)
					elabels = tuple(
						G[leaf][pattern[0]][el] for el in self.edge_labels
					)
					canonlist0.append(tuple((nlabels, elabels)))
				canonlist0.sort()
				canonlist0 = list(chain.from_iterable(canonlist0))
				canonlist3 = []
				for leaf in pattern[4:6]:
					nlabels = tuple(
						G.nodes[leaf][nl] for nl in self.node_labels
					)
					elabels = tuple(
						G[leaf][pattern[3]][el] for el in self.edge_labels
					)
					canonlist3.append(tuple((nlabels, elabels)))
				canonlist3.sort()
				canonlist3 = list(chain.from_iterable(canonlist3))

				# 2 possible key can be generated from 2 nodes with extended label 3,
				# select the one with lower lexicographic order.
				canonkey_t1 = tuple(
					['c']
					+ [tuple(
						G.nodes[pattern[0]][nl] for nl in self.node_labels
					)] + canonlist0
					+ [tuple(
						G.nodes[pattern[3]][nl] for nl in self.node_labels
					)]
					+ [tuple(
						G[pattern[3]][pattern[0]][el] for el in self.edge_labels
					)]
					+ canonlist3
					)
				canonkey_t2 = tuple(
					['c']
					+ [tuple(
						G.nodes[pattern[3]][nl] for nl in self.node_labels
					)] + canonlist3
					+ [tuple(
						G.nodes[pattern[0]][nl] for nl in self.node_labels
					)]
					+ [tuple(
						G[pattern[0]][pattern[3]][el] for el in self.edge_labels
					)]
					+ canonlist0
					)
				treelet.append(
					canonkey_t1 if canonkey_t1 < canonkey_t2 else canonkey_t2
				)
			canonkey_l.update(Counter(treelet))

			# pattern 9
			treelet = []
			for pattern in patterns['9']:
				canonkey2 = [
					tuple(G.nodes[pattern[4]][nl] for nl in self.node_labels),
					tuple(
						G[pattern[4]][pattern[2]][el] for el in self.edge_labels
					)]
				canonkey3 = [
					tuple(G.nodes[pattern[5]][nl] for nl in self.node_labels),
					tuple(
						G[pattern[5]][pattern[3]][el] for el in self.edge_labels
					)]
				prekey2 = [
					tuple(G.nodes[pattern[2]][nl] for nl in self.node_labels),
					tuple(
						G[pattern[2]][pattern[0]][el] for el in self.edge_labels
					)]
				prekey3 = [
					tuple(G.nodes[pattern[3]][nl] for nl in self.node_labels),
					tuple(
						G[pattern[3]][pattern[0]][el] for el in self.edge_labels
					)]
				if prekey2 + canonkey2 < prekey3 + canonkey3:
					canonkey_t = [tuple(
						G.nodes[pattern[1]][nl] for nl in self.node_labels
					)] \
					             + [tuple(
						G[pattern[1]][pattern[0]][el] for el in self.edge_labels
					)] \
					             + prekey2 + prekey3 + canonkey2 + canonkey3
				else:
					canonkey_t = [tuple(
						G.nodes[pattern[1]][nl] for nl in self.node_labels
					)] \
					             + [tuple(
						G[pattern[1]][pattern[0]][el] for el in self.edge_labels
					)] \
					             + prekey3 + prekey2 + canonkey3 + canonkey2
				treelet.append(
					tuple(
						['9']
						+ [tuple(
							G.nodes[pattern[0]][nl] for nl in self.node_labels
						)]
						+ canonkey_t
						)
				)
			canonkey_l.update(Counter(treelet))

			return canonkey_l

		return canonkey


	def _wrapper_get_canonkeys(self, itr_item):
		g = itr_item[0]
		i = itr_item[1]
		return i, self._get_canonkeys(g)


	def _add_dummy_labels(self, Gn=None):
		def _add_dummy(Gn):
			if len(self.node_labels) == 0 or (
					len(self.node_labels) == 1 and self.node_labels[
				0] == SpecialLabel.DUMMY):
				for i in range(len(Gn)):
					nx.set_node_attributes(Gn[i], '0', SpecialLabel.DUMMY)
				self.node_labels = [SpecialLabel.DUMMY]
			if len(self.edge_labels) == 0 or (
					len(self.edge_labels) == 1 and self.edge_labels[
				0] == SpecialLabel.DUMMY):
				for i in range(len(Gn)):
					nx.set_edge_attributes(Gn[i], '0', SpecialLabel.DUMMY)
				self.edge_labels = [SpecialLabel.DUMMY]


		if Gn is None or Gn is self._graphs:
			# Add dummy labels for the copy of self._graphs.
			try:
				check_is_fitted(self, ['_dummy_labels_considered'])
				if not self._dummy_labels_considered:
					Gn = self._graphs  # @todo: ?[g.copy() for g in self._graphs]
					_add_dummy(Gn)
					self._graphs = Gn
					self._dummy_labels_considered = True
			except NotFittedError:
				Gn = self._graphs  # @todo: ?[g.copy() for g in self._graphs]
				_add_dummy(Gn)
				self._graphs = Gn
				self._dummy_labels_considered = True

		else:
			# Add dummy labels for the input.
			_add_dummy(Gn)



# -*-coding:utf-8 -*-
"""gklearn - datasets module

Implement some methods to manage graph datasets
 graph_fetcher.py : fetch graph datasets from the Internet.


"""

# info
__version__ = "0.2"
__author__ = "Linlin Jia"
__date__ = "October 2020"


from gklearn.dataset.metadata import DATABASES, DATASET_META
from gklearn.dataset.metadata import GREYC_META, IAM_META, TUDataset_META
from gklearn.dataset.metadata import list_of_databases, list_of_datasets
from gklearn.dataset.graph_synthesizer import GraphSynthesizer
from gklearn.dataset.data_fetcher import DataFetcher
from gklearn.dataset.file_managers import DataLoader, DataSaver
from gklearn.dataset.dataset import Dataset, split_dataset_by_target